import sqlite3, json, shutil

# TODO
class SqliteDatabase:
    """
    functions which only meant to be used frequently are defined.

    if there is no functions you want to use, use ``getCursor`` instead
    """
    def __init__(self):
        super().__init__()
        self.__initVal()
        self.__initDb()
        self.__createImage()

    def __initVal(self):
        # db names
        self.__db_filename = 'image.db'

        # image table names
        self.__image_tb_nm = 'image_tb'
        self.__image_tb_tr_nm = 'image_tr'
        self.__image_unit_tb_nm = 'image_unit_tb'

        # info table names
        self.__dall_e_info_tb_nm = 'dall_e_info_tb'
        self.__stable_diffusion_info_tb_nm = 'stable_diffusion_tb'

        # model type (chat, etc.)
        self.__model_type = 1

        # DALL-E
        self.__image_dall_e_default_value = {
            'engine': "DALL-E",
            'n': 1,
            'width': 1024,
            'height': 1024,
            # 'response_format':
        }
        
        # SD
        self.__image_sd_default_value = {
            'height': 512,
            'width': 512,
            'steps': 30,
            'samples': 1,
            'cfg_scale': 7,
            'engine': 'stable-diffusion-xl-beta-v2-2-2',
            'sampler': 'k_dpmpp_2m',
            'seed': 0, # random
        }
        
        self.__each_info_dict = {1: [self.__dall_e_info_tb_nm, self.__image_dall_e_default_value],
                                 2: [self.__stable_diffusion_info_tb_nm, self.__image_sd_default_value], }

    def __initDb(self):
        try:
            # Connect to the database (create a new file if it doesn't exist)
            self.__conn = sqlite3.connect(self.__db_filename)
            self.__conn.execute('PRAGMA foreign_keys = ON;')
            self.__conn.commit()

            self.__c = self.__conn.cursor()
            self.__createInfo()
        except sqlite3.Error as e:
            print(f"An error occurred while connecting to the database: {e}")
            raise

    def __createDallE(self):
        # Check if the table exists
        self.__c.execute(f"SELECT count(*) FROM sqlite_master WHERE type='table' AND name='{self.__dall_e_info_tb_nm}'")
        if self.__c.fetchone()[0] == 1:
            # Check if each column already exists in the table
            existing_columns = [row[1] for row in self.__c.execute(f"PRAGMA table_info({self.__dall_e_info_tb_nm});")]
            new_columns = [col for col in self.__image_dall_e_default_value.keys() if col not in existing_columns]
            # Add the new columns if they don't already exist
            # TODO specify the type
            for col in new_columns:
                d_value = self.__image_dall_e_default_value[col]
                if isinstance(self.__image_dall_e_default_value[col], str):
                    d_value = f'"{d_value}"' if len(self.__image_dall_e_default_value[col].split()) > 0 else d_value
                self.__c.execute(f"ALTER TABLE {self.__dall_e_info_tb_nm} ADD COLUMN {col} DEFAULT {d_value}")
        else:
            self.__c.execute(f'''CREATE TABLE {self.__dall_e_info_tb_nm}
                                     (id INTEGER PRIMARY KEY,
                                      n TEXT DEFAULT '{self.__image_dall_e_default_value['n']}',
                                      width INTEGER DEFAULT {self.__image_dall_e_default_value['width']},
                                      height INTEGER DEFAULT {self.__image_dall_e_default_value['height']},
                                      update_dt DATETIME DEFAULT CURRENT_TIMESTAMP,
                                      insert_dt DATETIME DEFAULT CURRENT_TIMESTAMP)''')

            # Commit the transaction
            self.__conn.commit()

            # insert default record
            self.__c.execute(f'''INSERT INTO {self.__dall_e_info_tb_nm}
                                            (
                                                n,
                                                width,
                                                height
                                            ) VALUES
                                            (
                                                {','.join(['?' for _ in range(len(self.__image_dall_e_default_value))])}
                                            )
                                         ''', tuple(self.__image_dall_e_default_value.values()))

    def __createStableDiffusion(self):
        # Check if the table exists
        self.__c.execute(f"SELECT count(*) FROM sqlite_master WHERE type='table' AND name='{self.__stable_diffusion_info_tb_nm}'")
        if self.__c.fetchone()[0] == 1:
            pass
        else:
            self.__c.execute(f'''CREATE TABLE {self.__stable_diffusion_info_tb_nm}
                                             (
                                                id INTEGER PRIMARY KEY,
                                                height INTEGER DEFAULT '{self.__image_sd_default_value['height']},
                                                width INTEGER DEFAULT '{self.__image_sd_default_value['width']},
                                                steps INTEGER DEFAULT '{self.__image_sd_default_value['steps']},
                                                samples INTEGER DEFAULT '{self.__image_sd_default_value['samples']},
                                                cfg_scale INTEGER DEFAULT '{self.__image_sd_default_value['cfg_scale']},
                                                engine VARCHAR(50) DEFAULT '{self.__image_sd_default_value['engine']},
                                                sampler VARCHAR(50) DEFAULT '{self.__image_sd_default_value['sampler']},
                                                seed INTEGER DEFAULT '{self.__image_sd_default_value['seed']},
                                              update_dt DATETIME DEFAULT CURRENT_TIMESTAMP,
                                              insert_dt DATETIME DEFAULT CURRENT_TIMESTAMP)''')

            # Commit the transaction
            self.__conn.commit()

            # insert default record
            self.__c.execute(f'''INSERT INTO {self.__stable_diffusion_info_tb_nm}
                                                    (
                                                        height,
                                                        width,
                                                        steps,
                                                        samples,
                                                        cfg_scale,
                                                        engine,
                                                        sampler,
                                                        seed
                                                    ) VALUES
                                                    (
                                                        {','.join(['?' for _ in range(len(self.__image_sd_default_value))])}
                                                    )
                                                 ''', tuple(self.__image_sd_default_value.values()))

    def __createInfo(self):
        try:
            self.__createDallE()
            self.__createStableDiffusion()
            # Commit the transaction
            self.__conn.commit()
        except sqlite3.Error as e:
            print(f"An error occurred while creating the table: {e}")
            raise

    def __createImage(self):
        try:
            # Check if the table exists
            self.__c.execute(f"SELECT count(*) FROM sqlite_master WHERE type='table' AND name='{self.__image_tb_nm}'")
            if self.__c.fetchone()[0] == 1:
                # each image table already exists
                pass
            else:
                # Create a table with update_dt and insert_dt columns
                self.__c.execute(f'''CREATE TABLE {self.__image_tb_nm}
                             (id INTEGER PRIMARY KEY,
                              name TEXT,
                              update_dt DATETIME DEFAULT CURRENT_TIMESTAMP,
                              insert_dt DATETIME DEFAULT CURRENT_TIMESTAMP)''')
                # Create a trigger to update the update_dt column with the current timestamp
                self.__c.execute(f'''CREATE TRIGGER {self.__image_tb_tr_nm}
                             AFTER UPDATE ON {self.__image_tb_nm}
                             FOR EACH ROW
                             BEGIN
                               UPDATE {self.__image_tb_nm}
                               SET update_dt=CURRENT_TIMESTAMP
                               WHERE id=OLD.id;
                             END;''')
                # Commit the transaction
                self.__conn.commit()
        except sqlite3.Error as e:
            print(f"An error occurred while creating the table: {e}")
            raise

    def selectAllInfo(self):
        """
        select all info
        FIXME
        """
        try:
            # filter bool type fields
            bool_type_column = [row[1] for row in self.__c.execute(f'PRAGMA table_info({self.__dall_e_info_tb_nm})').fetchall() if row[2] == 'BOOL']

            # Execute the SELECT statement
            self.__c.execute(f'SELECT {",".join(list(self.__image_dall_e_default_value.keys()))} FROM {self.__dall_e_info_tb_nm}')

            # Get the column names
            column_names = [description[0] for description in self.__c.description]

            rows = self.__c.fetchall()

            info_dict_arr = []

            # Get the rows with field names and values
            for row in rows:
                info_dict = {}
                for i, value in enumerate(row):
                    if column_names[i] in bool_type_column:
                        value = True if value == 1 else False
                    info_dict[column_names[i]] = value
                info_dict_arr.append(info_dict)

            return info_dict_arr
        except sqlite3.Error as e:
            print(f"An error occurred while creating the table: {e}")
            raise

    def selectInfo(self, id=None):
        try:
            id = id if id else self.__model_type

            # filter bool type fields
            bool_type_column = [row[1] for row in self.__c.execute(f'PRAGMA table_info({self.__each_info_dict[id][0]})').fetchall() if row[2] == 'BOOL']

            # Execute the SELECT statement
            self.__c.execute(f'SELECT {",".join(list(self.__each_info_dict[id][1].keys()))} FROM {self.__each_info_dict[id][0]}')

            # Get the column names
            column_names = [description[0] for description in self.__c.description]

            row = self.__c.fetchone()

            info_dict = {}
            for i, value in enumerate(row):
                if column_names[i] in bool_type_column:
                    value = True if value == 1 else False
                info_dict[column_names[i]] = value

            return info_dict
        except sqlite3.Error as e:
            print(f"An error occurred while creating the table: {e}")
            raise

    def updateInfo(self, id, field, value):
        try:
            self.__c.execute(f'UPDATE {self.__each_info_dict[id][0]} SET {field}=(?) WHERE id=1', (value,))
            self.__conn.commit()
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def selectAllImage(self):
        """
        select all image
        """
        try:
            self.__c.execute(f'SELECT * FROM {self.__image_tb_nm}')
            return self.__c.fetchall()
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def selectImage(self, id):
        """
        select specific image
        """
        try:
            self.__c.execute(f'SELECT * FROM {self.__image_tb_nm} WHERE id={id}')
            return self.__c.fetchone()
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def insertImage(self, name):
        try:
            # Insert a row into the table
            self.__c.execute(f'INSERT INTO {self.__image_tb_nm} (name) VALUES (?)', (name,))
            new_id = self.__c.lastrowid
            # Commit the transaction
            self.__conn.commit()
            self.__createImageUnit(new_id)
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def updateImage(self, id, name):
        try:
            self.__c.execute(f'UPDATE {self.__image_tb_nm} SET name=(?) WHERE id={id}', (name,))
            self.__conn.commit()
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def deleteImage(self, id):
        try:
            self.__c.execute(f'DELETE FROM {self.__image_tb_nm} WHERE id={id}')
            self.__conn.commit()
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def __createImageUnit(self, id_fk):
        try:
            # Check if the table exists
            self.__c.execute(f"SELECT count(*) FROM sqlite_master WHERE type='table' AND name='{self.__image_unit_tb_nm}{id_fk}'")
            if self.__c.fetchone()[0] == 1:
                # each image table already exists
                pass
            else:
                self.__c.execute(f'''CREATE TABLE {self.__image_unit_tb_nm}{id_fk}
                                         (id INTEGER PRIMARY KEY,
                                          id_fk INTEGER,
                                          is_user INTEGER,
                                          image TEXT,
                                          update_dt DATETIME DEFAULT CURRENT_TIMESTAMP,
                                          insert_dt DATETIME DEFAULT CURRENT_TIMESTAMP,
                                          FOREIGN KEY (id_fk) REFERENCES {self.__image_tb_nm}(id) ON DELETE CASCADE)''')

                # insert trigger
                self.__c.execute(f'''
                    CREATE TRIGGER image_tb_updated_by_unit_inserted_tr{id_fk}
                    AFTER INSERT ON {self.__image_unit_tb_nm}{id_fk}
                    BEGIN
                      UPDATE {self.__image_tb_nm} SET update_dt = CURRENT_TIMESTAMP WHERE id = NEW.id_fk;
                    END
                ''')

                # update trigger
                self.__c.execute(f'''
                    CREATE TRIGGER image_tb_updated_by_unit_updated_tr{id_fk}
                    AFTER UPDATE ON {self.__image_unit_tb_nm}{id_fk}
                    BEGIN
                      UPDATE {self.__image_tb_nm} SET update_dt = CURRENT_TIMESTAMP WHERE id = NEW.id_fk;
                    END
                ''')

                # delete trigger
                self.__c.execute(f'''
                    CREATE TRIGGER image_tb_updated_by_unit_deleted_tr{id_fk}
                    AFTER DELETE ON {self.__image_unit_tb_nm}{id_fk}
                    BEGIN
                      UPDATE {self.__image_tb_nm} SET update_dt = CURRENT_TIMESTAMP WHERE id = OLD.id_fk;
                    END
                ''')
                # Commit the transaction
                self.__conn.commit()
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def selectImageUnit(self, id):
        self.__c.execute(f'SELECT * FROM {self.getImageUnitTableName()}{id}')
        return [elem[3] for elem in self.__c.fetchall()]

    def insertImageUnit(self, id, user_f, image):
        try:
            # Insert a row into the table
            self.__c.execute(
                f'INSERT INTO {self.__image_unit_tb_nm}{id} (id_fk, is_user, image) VALUES (?, ?, ?)',
                (id, user_f, image,))
            # Commit the transaction
            self.__conn.commit()
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def setModelType(self, model_type: int):
        """
        :param model_type: it starts from 1
        :return:
        """
        self.__model_type = model_type

    def export(self, ids, saved_filename):
        shutil.copy2(self.__db_filename, saved_filename)
        conn = sqlite3.connect(saved_filename)

        placeholders = ','.join('?' for _ in ids)
        cursor = conn.cursor()
        for i in range(len(ids)):
            delete_image_q = f"DELETE FROM {self.__image_tb_nm} WHERE id in ({placeholders})"
            cursor.execute(delete_image_q, ids)
            drop_image_unit_tb_q = f"DROP TABLE {self.__image_unit_tb_nm}{ids[i]}"
            cursor.execute(drop_image_unit_tb_q)
            conn.commit()

        conn.close()

    def imageertJsonIntoSql(self):
        try:
            # Read data from json
            with open('test/image_history.json', 'r') as f:
                data = json.load(f)
                for obj in list(data.values())[0]:
                    id, title, image_data = obj.values()
                    # start from 1 in sql unlike json which starts from 0
                    id += 1
                    self.insertImage(title)
                    for i in range(len(image_data)):
                        # Insert a row into the table
                        self.__c.execute(
                            f'INSERT INTO {self.__image_unit_tb_nm}{id} (id_fk, is_user, image) VALUES (?, ?, ?)',
                            (id, i % 2 == 0, image_data[i],))
                        # Commit the transaction
                        self.__conn.commit()
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def getCursor(self):
        return self.__c

    def getImageTableName(self):
        return self.__image_tb_nm

    def getImageUnitTableName(self):
        return self.__image_unit_tb_nm

    def close(self):
        self.__conn.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # Close the connection
        self.__conn.close()


import os
import openai, requests

from qtpy.QtGui import QGuiApplication, QFont, QIcon, QColor
from qtpy.QtWidgets import QMainWindow, QToolBar, QHBoxLayout, QDialog, QLineEdit, QPushButton, QWidgetAction, QSpinBox, QLabel, QWidget, QApplication, \
    QComboBox, QSizePolicy, QStackedWidget, QAction, QMenu, QSystemTrayIcon, \
    QMessageBox, QCheckBox
from qtpy.QtCore import Qt, QCoreApplication, QSettings

from pyqt_openai.aboutDialog import AboutDialog
from pyqt_openai.customizeDialog import CustomizeDialog
from pyqt_openai.svgButton import SvgButton
from pyqt_openai.image_gen_widget.imageGeneratingToolWidget import ImageGeneratingToolWidget
from pyqt_openai.openAiChatBotWidget import OpenAIChatBotWidget

QApplication.setAttribute(Qt.AA_EnableHighDpiScaling)
QCoreApplication.setAttribute(Qt.AA_UseHighDpiPixmaps)  # HighDPI support
# qt version should be above 5.14
# todo check the qt version with qtpy
if os.environ['QT_API'] == 'pyqt5' or os.environ['QT_API'] != 'pyside6':
    QGuiApplication.setHighDpiScaleFactorRoundingPolicy(Qt.HighDpiScaleFactorRoundingPolicy.PassThrough)

QApplication.setFont(QFont('Arial', 12))
QApplication.setWindowIcon(QIcon('ico/openai.svg'))


class MainWindow(QMainWindow):
    def __init__(self):
        super().__init__()
        self.__initVal()
        self.__initUi()

    def __initVal(self):
        self.__settings_struct = QSettings('pyqt_openai.ini', QSettings.IniFormat)

    def __initUi(self):
        self.setWindowTitle('PyQt OpenAI Chatbot')

        self.__openAiChatBotWidget = OpenAIChatBotWidget()
        self.__openAiChatBotWidget.notifierWidgetActivated.connect(self.show)
        self.__imageGeneratingToolWidget = ImageGeneratingToolWidget()
        self.__imageGeneratingToolWidget.notifierWidgetActivated.connect(self.show)
        self.__mainWidget = QStackedWidget()
        self.__mainWidget.addWidget(self.__openAiChatBotWidget)
        self.__mainWidget.addWidget(self.__imageGeneratingToolWidget)

        self.__setActions()
        self.__setMenuBar()
        self.__setTrayMenu()
        self.__setToolBar()

        # load ini file
        self.__loadApiKeyInIni()

        # check if loaded API_KEY from ini file is not empty
        if openai.api_key:
            self.__setApi()
        # if it is empty
        else:
            self.__setAIEnabled(False)
            self.__apiCheckPreviewLbl.hide()

        self.setCentralWidget(self.__mainWidget)
        self.resize(1024, 768)

    def __setActions(self):
        # menu action
        self.__exitAction = QAction("Exit", self)
        self.__exitAction.triggered.connect(self.close)

        self.__aboutAction = QAction("About...", self)
        self.__aboutAction.triggered.connect(self.__showAboutDialog)

        # toolbar action
        self.__chooseAiAction = QWidgetAction(self)
        self.__chooseAiCmbBox = QComboBox()
        self.__chooseAiCmbBox.addItems(['Chat', 'Image'])
        self.__chooseAiCmbBox.setSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.MinimumExpanding)
        self.__chooseAiCmbBox.currentIndexChanged.connect(self.__aiTypeChanged)
        self.__chooseAiAction.setDefaultWidget(self.__chooseAiCmbBox)

        self.__stackAction = QWidgetAction(self)
        self.__stackBtn = SvgButton()
        self.__stackBtn.setIcon('ico/stackontop.svg')
        self.__stackBtn.setCheckable(True)
        self.__stackBtn.toggled.connect(self.__stackToggle)
        self.__stackAction.setDefaultWidget(self.__stackBtn)
        self.__stackBtn.setToolTip('Always On Top')

        self.__customizeAction = QWidgetAction(self)
        self.__customizeBtn = SvgButton()
        self.__customizeBtn.setIcon('ico/customize.svg')
        self.__customizeBtn.clicked.connect(self.__executeCustomizeDialog)
        self.__customizeAction.setDefaultWidget(self.__customizeBtn)
        self.__customizeBtn.setToolTip('Customize (working)')

        self.__transparentAction = QWidgetAction(self)
        self.__transparentSpinBox = QSpinBox()
        self.__transparentSpinBox.setRange(0, 100)
        self.__transparentSpinBox.setValue(100)
        self.__transparentSpinBox.valueChanged.connect(self.__setTransparency)
        self.__transparentSpinBox.setToolTip('Set Transparency of Window')

        lay = QHBoxLayout()
        lay.addWidget(QLabel('Window Transparency'))
        lay.addWidget(self.__transparentSpinBox)

        transparencyActionWidget = QWidget()
        transparencyActionWidget.setLayout(lay)
        self.__transparentAction.setDefaultWidget(transparencyActionWidget)

        self.__showAiToolBarAction = QWidgetAction(self)
        self.__showAiToolBarChkBox = QCheckBox('Show AI Toolbar')
        self.__showAiToolBarChkBox.setChecked(True)
        self.__showAiToolBarChkBox.toggled.connect(self.__showAiToolBarChkBoxChecked)
        self.__showAiToolBarAction.setDefaultWidget(self.__showAiToolBarChkBox)

        self.__apiCheckPreviewLbl = QLabel()
        self.__apiCheckPreviewLbl.setFont(QFont('Arial', 10))

        apiLbl = QLabel('API')

        self.__apiLineEdit = QLineEdit()
        self.__apiLineEdit.setPlaceholderText('Write your API Key...')
        self.__apiLineEdit.returnPressed.connect(self.__setApi)
        self.__apiLineEdit.setEchoMode(QLineEdit.Password)

        apiBtn = QPushButton('Use')
        apiBtn.clicked.connect(self.__setApi)

        lay = QHBoxLayout()
        lay.addWidget(apiLbl)
        lay.addWidget(self.__apiLineEdit)
        lay.addWidget(apiBtn)
        lay.addWidget(self.__apiCheckPreviewLbl)
        lay.setContentsMargins(0, 0, 0, 0)

        apiWidget = QWidget()
        apiWidget.setLayout(lay)

        self.__apiAction = QWidgetAction(self)
        self.__apiAction.setDefaultWidget(apiWidget)

    def __setMenuBar(self):
        menubar = self.menuBar()

        # create the "File" menu
        fileMenu = QMenu("File", self)
        fileMenu.addAction(self.__exitAction)
        menubar.addMenu(fileMenu)

        # create the "Help" menu
        helpMenu = QMenu("Help", self)
        menubar.addMenu(helpMenu)

        helpMenu.addAction(self.__aboutAction)

    def __setTrayMenu(self):
        # background app
        menu = QMenu()

        action = QAction("Quit", self)
        action.setIcon(QIcon('ico/close.svg'))

        action.triggered.connect(app.quit)

        menu.addAction(action)

        tray_icon = QSystemTrayIcon(app)
        tray_icon.setIcon(QIcon('ico/openai.svg'))
        tray_icon.activated.connect(self.__activated)

        tray_icon.setContextMenu(menu)

        tray_icon.show()

    def __activated(self, reason):
        if reason == 3:
            self.show()

    def __setToolBar(self):
        aiTypeToolBar = QToolBar()
        aiTypeToolBar.setMovable(False)
        aiTypeToolBar.addAction(self.__chooseAiAction)

        windowToolBar = QToolBar()
        lay = QHBoxLayout()
        windowToolBar.addAction(self.__stackAction)
        windowToolBar.addAction(self.__customizeAction)
        windowToolBar.addAction(self.__transparentAction)
        windowToolBar.addAction(self.__showAiToolBarAction)
        windowToolBar.addAction(self.__apiAction)
        windowToolBar.setLayout(lay)
        windowToolBar.setMovable(False)
        self.addToolBar(aiTypeToolBar)
        self.addToolBar(windowToolBar)
        # QToolbar's layout can't be set spacing with lay.setSpacing so i've just did this instead
        windowToolBar.setStyleSheet('QToolBar { spacing: 2px; }')

    def __setApiKey(self, api_key):
        # for script
        openai.api_key = api_key
        # for subprocess (mostly)
        os.environ['OPENAI_API_KEY'] = api_key
        # for showing to the user
        self.__apiLineEdit.setText(api_key)

    def __loadApiKeyInIni(self):
        # this api key should be yours
        if self.__settings_struct.contains('API_KEY'):
            self.__setApiKey(self.__settings_struct.value('API_KEY'))
        else:
            self.__settings_struct.setValue('API_KEY', '')

    def __setAIEnabled(self, f):
        self.__openAiChatBotWidget.setAIEnabled(f)
        self.__imageGeneratingToolWidget.setAIEnabled(f)

    def __setApi(self):
        try:
            api_key = self.__apiLineEdit.text()
            response = requests.get('https://api.openai.com/v1/engines', headers={'Authorization': f'Bearer {api_key}'})
            f = response.status_code == 200
            self.__setAIEnabled(f)
            if f:
                self.__setApiKey(api_key)
                self.__settings_struct.setValue('API_KEY', api_key)

                self.__openAiChatBotWidget.setModelInfoByModel(True)

                self.__apiCheckPreviewLbl.setStyleSheet("color: {}".format(QColor(0, 200, 0).name()))
                self.__apiCheckPreviewLbl.setText('API key is valid')
            else:
                raise Exception
        except Exception as e:
            self.__apiCheckPreviewLbl.setStyleSheet("color: {}".format(QColor(255, 0, 0).name()))
            self.__apiCheckPreviewLbl.setText('API key is invalid')
            self.__setAIEnabled(False)
            print(e)
        finally:
            self.__apiCheckPreviewLbl.show()

    def __showAboutDialog(self):
        aboutDialog = AboutDialog()
        aboutDialog.exec()

    def __stackToggle(self, f):
        if f:
            self.setWindowFlags(self.windowFlags() | Qt.WindowStaysOnTopHint)
        else:
            self.setWindowFlags(self.windowFlags() & ~Qt.WindowStaysOnTopHint)
        self.show()

    def __setTransparency(self, v):
        self.setWindowOpacity(v / 100)

    def __showAiToolBarChkBoxChecked(self, f):
        self.__mainWidget.currentWidget().showAiToolBar(f)

    def __executeCustomizeDialog(self):
        dialog = CustomizeDialog(self)
        reply = dialog.exec()
        if reply == QDialog.Accepted:
            pass

    def __aiTypeChanged(self, i):
        self.__mainWidget.setCurrentIndex(i)

    def closeEvent(self, e):
        message = 'The window has been closed. Would you like to continue running this app in the background?'
        closeMessageBox = QMessageBox(self)
        closeMessageBox.setWindowTitle('Wait!')
        closeMessageBox.setText(message)
        closeMessageBox.setStandardButtons(QMessageBox.Yes | QMessageBox.No)
        reply = closeMessageBox.exec()
        # Yes
        if reply == 16384:
            e.accept()
        # No
        elif reply == 65536:
            app.quit()
        return super().closeEvent(e)


if __name__ == "__main__":
    import sys

    app = QApplication(sys.argv)
    app.setQuitOnLastWindowClosed(False)
    w = MainWindow()
    w.show()
    sys.exit(app.exec())


from qtpy.QtWidgets import QWidget, QComboBox, QTextEdit, QLabel, QVBoxLayout, QApplication, QCheckBox, QFormLayout, \
    QSpinBox, QScrollArea

from pyqt_openai.sqlite import SqliteDatabase


class ImagePage(QScrollArea):
    def __init__(self, db: SqliteDatabase):
        super().__init__()
        self.__initVal(db)
        self.__initUi()

    def __initVal(self, db):
        self.__db = db
        self.__info_dict = self.__db.selectInfo(3)

    def __initUi(self):
        modelCmbBox = QComboBox()
        modelCmbBox.addItems(['DALL-E'])
        # modelCmbBox.addItems(['DALL-E', 'Midjourney', 'Stable Diffusion'])
        nSpinBox = QSpinBox()
        nSpinBox.setRange(1, 10)
        nSpinBox.setValue(self.__info_dict['n'])
        nSpinBox.valueChanged.connect(self.__nChanged)
        sizeCmbBox = QComboBox()
        sizeCmbBox.addItems(['256x256', '512x512', '1024x1024'])
        sizeCmbBox.setCurrentText(f"{self.__info_dict['width']}x{self.__info_dict['height']}")
        sizeCmbBox.currentTextChanged.connect(self.__sizeChanged)
        lay = QFormLayout()
        lay.addRow('Model', modelCmbBox)
        lay.addRow('Total', nSpinBox)
        lay.addRow('Size', sizeCmbBox)

        mainWidget = QWidget()
        mainWidget.setLayout(lay)

        self.setWidget(mainWidget)
        self.setWidgetResizable(True)

    def __nChanged(self, v):
        self.__db.updateInfo(3, 'n', v)

    def __sizeChanged(self, v):
        width, height = v.split('x')
        self.__db.updateInfo(3, 'width', width)
        self.__db.updateInfo(3, 'height', height)



import os
import warnings

import stability_sdk.interfaces.gooseai.generation.generation_pb2 as generation
from qtpy.QtCore import Signal, QThread, QSettings
from qtpy.QtWidgets import QWidget, QLineEdit, QVBoxLayout, QHBoxLayout, QGroupBox, QPushButton, QFormLayout, QSpinBox, QComboBox, \
    QLabel, QPlainTextEdit
from stability_sdk import client

from pyqt_openai.notifier import NotifierWidget
from pyqt_openai.svgLabel import SvgLabel

# Our Host URL should not be prepended with "https" nor should it have a trailing slash.
os.environ['STABILITY_HOST'] = 'grpc.stability.ai:443'


class StableDiffusionThread(QThread):
    replyGenerated = Signal(bytes)

    def __init__(self, answers, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__answers = answers

    def run(self):
        for resp in self.__answers:
            for artifact in resp.artifacts:
                if artifact.finish_reason == generation.FILTER:
                    warnings.warn(
                        "Your request activated the API's safety filters and could not be processed."
                        "Please modify the prompt and try again.")
                if artifact.type == generation.ARTIFACT_IMAGE:
                    self.replyGenerated.emit(artifact.binary)


class ImageStableDiffusionPage(QWidget):
    submitSd = Signal(bytes)
    notifierWidgetActivated = Signal()

    def __init__(self):
        super().__init__()
        self.__initVal()
        self.__initUi()

    def __initVal(self):
        self.__settings_struct = QSettings('pyqt_openai.ini', QSettings.IniFormat)

        self.__sampler_dict = {
            'SAMPLER_DDIM': generation.SAMPLER_DDIM,
            'SAMPLER_DDPM': generation.SAMPLER_DDPM,
            'SAMPLER_K_EULER': generation.SAMPLER_K_EULER,
            'SAMPLER_K_EULER_ANCESTRAL': generation.SAMPLER_K_EULER_ANCESTRAL,
            'SAMPLER_K_HEUN': generation.SAMPLER_K_HEUN,
            'SAMPLER_K_DPM_2': generation.SAMPLER_K_DPM_2,
            'SAMPLER_K_DPM_2_ANCESTRAL': generation.SAMPLER_K_DPM_2_ANCESTRAL,
            'SAMPLER_K_LMS': generation.SAMPLER_K_LMS,
            'SAMPLER_K_DPMPP_2S_ANCESTRAL': generation.SAMPLER_K_DPMPP_2S_ANCESTRAL,
            'SAMPLER_K_DPMPP_2M': generation.SAMPLER_K_DPMPP_2M,
            'SAMPLER_K_DPMPP_SDE': generation.SAMPLER_K_DPMPP_SDE,
        }

        self.__info_dict = {
            'height': '512',
            'width': '512',
            'steps': 30,
            'samples': 1,
            'cfg_scale': 7,
            'engine': 'stable-diffusion-xl-beta-v2-2-2',
            'sampler': 'k_dpmpp_2m',
            'seed': 0, # random
        }

    def __initUi(self):
        sdApiGrpBox = QGroupBox()
        sdApiGrpBox.setTitle('DreamStudio API')

        self.__apiLineEdit = QLineEdit()
        self.__apiLineEdit.setPlaceholderText('Write your DreamStudio API Key...')
        self.__apiLineEdit.textChanged.connect(self.__enableSubmitBtn)
        self.__apiLineEdit.setEchoMode(QLineEdit.Password)

        whatIsDreamStudioLbl = SvgLabel()
        whatIsDreamStudioLbl.setSvgFile('ico/help.svg')
        whatIsDreamStudioLbl.setToolTip('DreamStudio is a service which make it enable anyone to access the image generation tool\n without the need for software installation, coding knowledge, or a heavy-duty local GPU.')

        lay = QHBoxLayout()
        lay.addWidget(self.__apiLineEdit)
        lay.addWidget(whatIsDreamStudioLbl)
        lay.setContentsMargins(0, 0, 0, 0)
        sdApiInputWidget = QWidget()
        sdApiInputWidget.setLayout(lay)

        apiNoteLbl = QLabel('※ The validity of the API will be checked when you submit the form below.')

        lay = QVBoxLayout()
        lay.addWidget(sdApiInputWidget)
        lay.addWidget(apiNoteLbl)
        sdApiGrpBox.setLayout(lay)

        self.__modelCmbBox = QComboBox()
        self.__modelCmbBox.addItems([
            'stable-diffusion-v1',
            'stable-diffusion-v1-5',
            'stable-diffusion-512-v2-0',
            'stable-diffusion-768-v2-0',
            'stable-diffusion-512-v2-1',
            'stable-diffusion-768-v2-1',
            'stable-diffusion-xl-beta-v2-2-2',
            'stable-inpainting-v1-0',
            'stable-inpainting-512-v2-0',
        ])
        self.__modelCmbBox.setCurrentText(self.__info_dict['engine'])

        self.__upscaleEngineCmbBox = QComboBox()
        self.__upscaleEngineCmbBox.addItems([
            'esrgan-v1-x2plus', # cheap one
            'stable-diffusion-x4-latent-upscaler',
        ])

        self.__samplesSpinBox = QSpinBox()
        self.__samplesSpinBox.setRange(1, 10)
        self.__samplesSpinBox.setValue(self.__info_dict['samples'])
        # self.__samplesSpinBox.valueChanged.connect(self.__nChanged)

        self.__widthCmbBox = QComboBox()
        self.__widthCmbBox.addItems(['512', '768'])
        self.__widthCmbBox.setCurrentText(self.__info_dict['width'])

        self.__heightCmbBox = QComboBox()
        self.__heightCmbBox.addItems(['512', '768'])
        self.__heightCmbBox.setCurrentText(self.__info_dict['height'])

        self.__stepsSpinBox = QSpinBox()
        self.__stepsSpinBox.setRange(10, 150)
        self.__stepsSpinBox.setValue(self.__info_dict['steps'])

        self.__seedSpinBox = QSpinBox()
        self.__seedSpinBox.setValue(self.__info_dict['seed'])

        self.__cfgScalesSpinBox = QSpinBox()
        self.__cfgScalesSpinBox.setRange(1, 35)
        self.__cfgScalesSpinBox.setValue(self.__info_dict['cfg_scale'])

        self.__samplerCmbBox = QComboBox()
        self.__samplerCmbBox.addItems(list(self.__sampler_dict.keys()))
        self.__samplerCmbBox.setCurrentText(self.__info_dict['sampler'])

        self.__promptWidget = QPlainTextEdit()
        self.__submitBtn = QPushButton('Submit')
        self.__submitBtn.clicked.connect(self.__submit)
        self.__submitBtn.setEnabled(False)

        lay = QFormLayout()
        lay.addRow('Model', self.__modelCmbBox)
        lay.addRow('Samples', self.__samplesSpinBox)
        lay.addRow('Width', self.__widthCmbBox)
        lay.addRow('Height', self.__heightCmbBox)
        lay.addRow('Steps', self.__stepsSpinBox)
        lay.addRow('Seed', self.__seedSpinBox)
        lay.addRow('CFG scales', self.__cfgScalesSpinBox)
        lay.addRow('Sampler', self.__samplerCmbBox)
        lay.addRow(QLabel('Prompt'))
        lay.addRow(self.__promptWidget)

        generalGrpBox = QGroupBox('General')
        generalGrpBox.setLayout(lay)

        lay = QFormLayout()
        lay.addRow('Upscaling Model', self.__upscaleEngineCmbBox)
        advancedGrpBox = QGroupBox('Advanced (Working)')
        advancedGrpBox.setLayout(lay)
        advancedGrpBox.setEnabled(False)

        lay = QVBoxLayout()
        lay.addWidget(sdApiGrpBox)
        lay.addWidget(generalGrpBox)
        lay.addWidget(advancedGrpBox)
        lay.addWidget(self.__submitBtn)

        self.setLayout(lay)

        self.__loadApiKeyInIni()

    def __enableSubmitBtn(self):
        f = self.__apiLineEdit.text().strip() != ''
        self.__submitBtn.setEnabled(self.__apiLineEdit.text().strip() != '')
        if f:
            self.__settings_struct.setValue('SD_API_KEY', self.__apiLineEdit.text().strip())

    def __loadApiKeyInIni(self):
        # this api key should be yours
        if self.__settings_struct.contains('SD_API_KEY'):
            self.__apiLineEdit.setText(self.__settings_struct.value('SD_API_KEY'))
        else:
            self.__settings_struct.setValue('SD_API_KEY', '')

    def __submit(self):
        stability_api = client.StabilityInference(
            key=self.__apiLineEdit.text(),  # API Key reference.
            # upscale_engine=self.__upscaleEngineCmbBox.currentText(),
            verbose=True,  # Print debug messages.
            engine=self.__modelCmbBox.currentText(),  # Set the engine to use for generation.
            # Available engines: stable-diffusion-v1 stable-diffusion-v1-5 stable-diffusion-512-v2-0 stable-diffusion-768-v2-0
            # stable-diffusion-512-v2-1 stable-diffusion-768-v2-1 stable-diffusion-xl-beta-v2-2-2 stable-inpainting-v1-0 stable-inpainting-512-v2-0
        )

        # Set up our initial generation parameters.
        answers = stability_api.generate(
            prompt=self.__promptWidget.toPlainText(),
            seed=self.__seedSpinBox.value(),  # If a seed is provided, the resulting generated image will be deterministic.
            # What this means is that as long as all generation parameters remain the same, you can always recall the same image simply by generating it again.
            # Note: This isn't quite the case for CLIP Guided generations, which we tackle in the CLIP Guidance documentation.
            steps=self.__stepsSpinBox.value(),  # Amount of inference steps performed on image generation. Defaults to 30.
            cfg_scale=self.__cfgScalesSpinBox.value(),  # Influences how strongly your generation is guided to match your prompt.
            # Setting this value higher increases the strength in which it tries to match your prompt.
            # Defaults to 7.0 if not specified.
            width=int(self.__widthCmbBox.currentText()),  # Generation width, defaults to 512 if not included.
            height=int(self.__heightCmbBox.currentText()),  # Generation height, defaults to 512 if not included.
            samples=self.__samplesSpinBox.value(),  # Number of images to generate, defaults to 1 if not included.
            sampler=self.__samplerCmbBox.currentText()  # Choose which sampler we want to denoise our generation with.
            # Defaults to k_dpmpp_2m if not specified. Clip Guidance only supports ancestral samplers.
            # (Available Samplers: ddim, plms, k_euler, k_euler_ancestral, k_heun, k_dpm_2, k_dpm_2_ancestral, k_dpmpp_2s_ancestral, k_lms, k_dpmpp_2m, k_dpmpp_sde)
        )

        self.__t = StableDiffusionThread(answers)
        self.__submitBtn.setEnabled(False)
        self.__t.start()
        self.__t.replyGenerated.connect(self.__afterGenerated)

    def __afterGenerated(self, image_bin):
        self.submitSd.emit(image_bin)
        if not self.isVisible():
            self.__notifierWidget = NotifierWidget(informative_text='Response 👌', detailed_text='Click this!')
            self.__notifierWidget.show()
            self.__notifierWidget.doubleClicked.connect(self.notifierWidgetActivated)
        self.__submitBtn.setEnabled(True)



import openai

from qtpy.QtWidgets import QWidget, QPushButton, QComboBox, QPlainTextEdit, QSpinBox, QFormLayout, QTextEdit, QLabel
from qtpy.QtCore import Signal, QThread

from pyqt_openai.notifier import NotifierWidget


class DallEThread(QThread):
    replyGenerated = Signal(str)

    def __init__(self, openai_arg, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__openai_arg = openai_arg

    def run(self):
        response = openai.Image.create(
            **self.__openai_arg
        )

        for image_data in response['data']:
            image_url = image_data['url']
            self.replyGenerated.emit(image_url)


class ImageDallEPage(QWidget):
    submitDallE = Signal(str)
    notifierWidgetActivated = Signal()
    # class name: ImageDatabase
    # file name: imageDb

    def __init__(self):
        super().__init__()
        # self.__initVal()
        self.__initUi()

    def __initUi(self):
        self.__nSpinBox = QSpinBox()
        self.__nSpinBox.setRange(1, 10)
        # self.__nSpinBox.setValue(self.__info_dict['n'])
        # self.__nSpinBox.valueChanged.connect(self.__nChanged)
        self.__sizeCmbBox = QComboBox()
        self.__sizeCmbBox.addItems(['256x256', '512x512', '1024x1024'])
        # self.__sizeCmbBox.setCurrentText(f"{self.__info_dict['width']}x{self.__info_dict['height']}")
        self.__sizeCmbBox.currentTextChanged.connect(self.__sizeChanged)

        self.__promptWidget = QPlainTextEdit()
        self.__submitBtn = QPushButton('Submit')
        self.__submitBtn.clicked.connect(self.__submit)

        lay = QFormLayout()
        lay.addRow('Total', self.__nSpinBox)
        lay.addRow('Size', self.__sizeCmbBox)
        lay.addRow(QLabel('Prompt'))
        lay.addRow(self.__promptWidget)
        lay.addRow(self.__submitBtn)

        self.setLayout(lay)

    def __nChanged(self, v):
        pass
        # self.__db.updateInfo(3, 'n', v)

    def __sizeChanged(self, v):
        width, height = v.split('x')
        # self.__db.updateInfo(3, 'width', width)
        # self.__db.updateInfo(3, 'height', height)

    def __submit(self):
        # openai_arg = {
        #     "prompt": self.__prompt.getContent(),
        #     "n": info_dict['n'],
        #     "size": f"{info_dict['width']}x{info_dict['height']}"
        # }
        openai_arg = {
            "prompt": self.__promptWidget.toPlainText(),
            "n": self.__nSpinBox.value(),
            "size": self.__sizeCmbBox.currentText()
        }
        self.__t = DallEThread(openai_arg)
        self.__submitBtn.setEnabled(False)
        self.__t.start()
        self.__t.replyGenerated.connect(self.__afterGenerated)

    def __afterGenerated(self, image_url):
        self.submitDallE.emit(image_url)
        if not self.isVisible():
            self.__notifierWidget = NotifierWidget(informative_text='Response 👌', detailed_text='Click this!')
            self.__notifierWidget.show()
            self.__notifierWidget.doubleClicked.connect(self.notifierWidgetActivated)
        self.__submitBtn.setEnabled(True)



# save with json
# # make a new conversation unit
# def makeNewConvJson():
#     # Open the JSON file and load its contents into a Python dictionary
#     with open('conv_history.json', 'r') as f:
#         data = json.load(f)
#
#     with open('conv_history.json', 'w') as f:
#         lst = data['each_conv_lst']
#         max_id = max(lst, key=lambda x: x["id"])["id"]+1
#         data['each_conv_lst'].append({ 'id': max_id, 'title': 'New Chat', 'conv_data': [] })
#         f.write(json.dumps(data) + '\n')
#
# # update conversation unit
# def updateConvJson(id, title=None, conv_unit=None):
#     # Open the JSON file and load its contents into a Python dictionary
#     with open('conv_history.json', 'r') as f:
#         data = json.load(f)
#
#     with open('conv_history.json', 'w') as f:
#         lst = data['each_conv_lst']
#         obj = list(filter(lambda x: x["id"] == 1, lst))[0]
#         if title:
#             obj['title'] = title
#         if conv_unit:
#             obj['conv_data'].append(conv_unit)
#         json.dump(data, f)
#
# makeNewConvJson()
# updateConvJson(1, 'New Chat')
# updateConvJson(1, conv_unit=conv)

# hyperlink
# from pygments import highlight
# from pygments.lexers import PythonLexer
# from pygments.formatters import HtmlFormatter
#
# code = """
# def hello_world():
#     print("Hello, world!")
# """
#
# lexer = PythonLexer()
# formatter = HtmlFormatter(style='colorful')
#
# css_styles = formatter.get_style_defs('.highlight')
#
# html_code = f"""
# <html>
#     <head>
#         <style>
#             {css_styles}
#         </style>
#     </head>
#     <body>
#         {highlight(code, lexer, formatter)}
#     </body>
# </html>
# """
#
# print(html_code)
#


from qtpy.QtWidgets import QWidget, QTabWidget, QVBoxLayout
from qtpy.QtCore import Signal

from pyqt_openai.image_gen_widget.imageDallEPage import ImageDallEPage
from pyqt_openai.image_gen_widget.imageStableDiffusionPage import ImageStableDiffusionPage


class RightSideBar(QWidget):
    submitDallE = Signal(str)
    submitSd = Signal(bytes)
    notifierWidgetActivated = Signal()

    def __init__(self):
        super().__init__()
        self.__initUi()

    def __initUi(self):
        w1 = ImageDallEPage()
        w1.notifierWidgetActivated.connect(self.notifierWidgetActivated)
        w2 = ImageStableDiffusionPage()
        w2.notifierWidgetActivated.connect(self.notifierWidgetActivated)

        w1.submitDallE.connect(self.submitDallE)
        w2.submitSd.connect(self.submitSd)

        tabWidget = QTabWidget()
        tabWidget.addTab(w1, 'DALL-E')
        tabWidget.addTab(w2, 'Stable Diffusion')
        lay = QVBoxLayout()
        lay.addWidget(tabWidget)
        lay.setContentsMargins(0, 0, 0, 0)
        self.setLayout(lay)


import pyperclip
from qtpy.QtWidgets import QTextBrowser, QWidget, QLabel, QVBoxLayout, QPushButton, QTabWidget, QScrollArea

from pyqt_openai.prompt.propPage import PropPage
from pyqt_openai.prompt.templatePage import TemplatePage
from pyqt_openai.sqlite import SqliteDatabase


class PromptGeneratorWidget(QScrollArea):
    def __init__(self, db: SqliteDatabase):
        super().__init__()
        self.__initVal(db)
        self.__initUi()

    def __initVal(self, db):
        self.__db = db

    def __initUi(self):
        propmtLbl = QLabel('Prompt')

        propPage = PropPage(self.__db)
        propPage.updated.connect(self.__textChanged)

        templatePage = TemplatePage(self.__db)
        templatePage.updated.connect(self.__textChanged)

        self.__prompt = QTextBrowser()
        self.__prompt.setPlaceholderText('Generated Prompt')

        promptTabWidget = QTabWidget()
        promptTabWidget.addTab(propPage, 'Property')
        promptTabWidget.addTab(templatePage, 'Template')

        previewLbl = QLabel('Preview')

        copyBtn = QPushButton('Copy')
        copyBtn.clicked.connect(self.__copy)

        lay = QVBoxLayout()
        lay.addWidget(propmtLbl)
        lay.addWidget(promptTabWidget)
        lay.addWidget(previewLbl)
        lay.addWidget(self.__prompt)
        lay.addWidget(copyBtn)

        mainWidget = QWidget()
        mainWidget.setLayout(lay)

        self.setWidget(mainWidget)
        self.setWidgetResizable(True)

        self.setStyleSheet('QScrollArea { border: 0 }')

    def __textChanged(self, prompt_text):
        self.__prompt.clear()
        self.__prompt.setText(prompt_text)

    def __copy(self):
        pyperclip.copy(self.__prompt.toPlainText())


from qtpy.QtWidgets import QDialog, QVBoxLayout, QLineEdit, QFrame, QPushButton, QHBoxLayout, QWidget
from qtpy.QtCore import Qt
from qtpy.QtGui import QIcon


class InputDialog(QDialog):
    def __init__(self, title, text, parent=None):
        super().__init__(parent)
        self.__initUi(title, text)

    def __initUi(self, title, text):
        self.setWindowTitle(title)
        self.setWindowFlags(Qt.Window | Qt.WindowCloseButtonHint)

        self.__newName = QLineEdit(text)
        self.__newName.textChanged.connect(self.__setAccept)

        sep = QFrame()
        sep.setFrameShape(QFrame.HLine)
        sep.setFrameShadow(QFrame.Sunken)

        self.__okBtn = QPushButton('OK')
        self.__okBtn.clicked.connect(self.accept)

        cancelBtn = QPushButton('Cancel')
        cancelBtn.clicked.connect(self.close)

        lay = QHBoxLayout()
        lay.addWidget(self.__okBtn)
        lay.addWidget(cancelBtn)
        lay.setAlignment(Qt.AlignRight)
        lay.setContentsMargins(0, 0, 0, 0)

        okCancelWidget = QWidget()
        okCancelWidget.setLayout(lay)

        lay = QVBoxLayout()
        lay.addWidget(self.__newName)
        lay.addWidget(sep)
        lay.addWidget(okCancelWidget)

        self.setLayout(lay)

    def getText(self):
        return self.__newName.text()

    def __setAccept(self, text):
        self.__okBtn.setEnabled(text.strip() != '')


from qtpy.QtGui import QPixmap
from qtpy.QtWidgets import QGridLayout, QWidget, QScrollArea, \
    QSizePolicy
from qtpy.QtCore import Qt, Signal

from pyqt_openai.image_gen_widget.thumbnailView import ThumbnailView


class ExplorerWidget(QScrollArea):
    clicked = Signal(QPixmap)

    def __init__(self):
        super().__init__()
        self.__initUi()

    def __initUi(self):
        lay = QGridLayout()
        lay.setAlignment(Qt.AlignLeft | Qt.AlignTop)

        mainWidget = QWidget()
        mainWidget.setLayout(lay)

        self.setWidget(mainWidget)
        self.setWidgetResizable(True)

    def insertWidgetAsFirst(self, widget):
        lay = self.widget().layout()

        checkout = 6
        for i in range(lay.count()):
            r_idx, c_idx = divmod(i, checkout)
            item = lay.itemAtPosition(r_idx, c_idx)
            if item is not None:
                if checkout-c_idx == 1:
                    r_idx += 1
                    c_idx = -1
                lay.addWidget(item.widget(), r_idx, c_idx+1)
                lay.setColumnStretch(c_idx+1, 1)

        # Insert the new widget in the top-left corner
        lay.addWidget(widget, 0, 0)

    def __addThumbnail(self):
        thumbnail = ThumbnailView()
        thumbnail.clicked.connect(self.clicked)
        thumbnail.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.insertWidgetAsFirst(thumbnail)
        return thumbnail

    def addContent(self, content):
        thumbnail = self.__addThumbnail()
        thumbnail.setContent(content)

    def addFilename(self, filename):
        thumbnail = self.__addThumbnail()
        thumbnail.setFilename(filename)



from qtpy.QtWidgets import QWidget, QComboBox, QTextEdit, QLabel, QVBoxLayout, QApplication, QCheckBox, QDoubleSpinBox, \
    QSpinBox, QPushButton

from pyqt_openai.apiData import getChatModel
from pyqt_openai.sqlite import SqliteDatabase


class ChatPage(QWidget):
    def __init__(self, db: SqliteDatabase, ini_etc_dict):
        super().__init__()
        self.__initVal(db, ini_etc_dict)
        self.__initUi()

    def __setChatInfo(self, db):
        self.__db = db
        self.__info_dict = self.__db.selectInfo(1)

        # # set each field as variable since these are being used a lot
        # engine = info_dict['engine']
        # temperature = info_dict['temperature']
        # max_tokens = info_dict['max_tokens']
        # top_p = info_dict['top_p']
        # frequency_penalty = info_dict['frequency_penalty']
        # presence_penalty = info_dict['presence_penalty']
        # stream = info_dict['stream']

    def __initVal(self, db, ini_etc_dict):
        self.__setChatInfo(db)

        # set each field as variable since these are being used a lot
        self.__stream = self.__info_dict.get('stream', False)

        self.__ini_etc_dict = ini_etc_dict

    def __initUi(self):
        systemlbl = QLabel('System')
        self.__systemTextEdit = QTextEdit()
        self.__systemTextEdit.setText('You are a helpful assistant.')
        saveSystemBtn = QPushButton('Save System')
        saveSystemBtn.clicked.connect(self.__saveSystem)
        modelCmbBox = QComboBox()
        modelCmbBox.addItems(getChatModel())
        modelCmbBox.currentTextChanged.connect(self.__modelChanged)

        # temperatureSpinBox = QDoubleSpinBox()
        # temperatureSpinBox.setRange(0, 1)
        # temperatureSpinBox.setAccelerated(True)
        # temperatureSpinBox.setSingleStep(0.01)
        # # temperatureSpinBox.setValue(self.__info_dict['temperature'])
        # # temperatureSpinBox.valueChanged.connect(self.__temperatureChanged)
        #
        # # later
        # # maxTokensSpinBox = QSpinBox()
        # # maxTokensSpinBox.setRange(0, 4000)
        # # maxTokensSpinBox.setAccelerated(True)
        # # maxTokensSpinBox.setValue(self.__info_dict['max_tokens'])
        # # maxTokensSpinBox.valueChanged.connect(self.__maxTokensChanged)
        #
        # toppSpinBox = QDoubleSpinBox()
        # toppSpinBox.setRange(0, 1)
        # toppSpinBox.setAccelerated(True)
        # toppSpinBox.setSingleStep(0.01)
        # # toppSpinBox.setValue(self.__info_dict['top_p'])
        # # toppSpinBox.valueChanged.connect(self.__toppChanged)
        #
        # frequencyPenaltySpinBox = QDoubleSpinBox()
        # frequencyPenaltySpinBox.setRange(0, 2)
        # frequencyPenaltySpinBox.setAccelerated(True)
        # frequencyPenaltySpinBox.setSingleStep(0.01)
        # # frequencyPenaltySpinBox.setValue(self.__info_dict['frequency_penalty'])
        # # frequencyPenaltySpinBox.valueChanged.connect(self.__frequencyPenaltyChanged)
        #
        # presencePenaltySpinBox = QDoubleSpinBox()
        # presencePenaltySpinBox.setRange(0, 2)
        # presencePenaltySpinBox.setAccelerated(True)
        # presencePenaltySpinBox.setSingleStep(0.01)
        # # presencePenaltySpinBox.setValue(self.__info_dict['presence_penalty'])
        # # presencePenaltySpinBox.valueChanged.connect(self.__presencePenaltyChanged)

        streamChkBox = QCheckBox()
        streamChkBox.setChecked(self.__stream)
        streamChkBox.toggled.connect(self.__streamChecked)
        streamChkBox.setText('Stream')

        finishReasonChkBox = QCheckBox('Show Finish Reason (working)')
        finishReasonChkBox.setChecked(self.__ini_etc_dict['finishReason'])
        finishReasonChkBox.toggled.connect(self.__finishReasonChecked)
        finishReasonChkBox.setText('Show Finish Reason')

        lay = QVBoxLayout()
        lay.addWidget(systemlbl)
        lay.addWidget(self.__systemTextEdit)
        lay.addWidget(saveSystemBtn)
        lay.addWidget(modelCmbBox)
        lay.addWidget(streamChkBox)
        lay.addWidget(finishReasonChkBox)

        self.setLayout(lay)

    def __saveSystem(self):
        self.__info_dict['system'] = self.__systemTextEdit.toPlainText()
        self.__db.updateInfo(1, 'system', self.__info_dict['system'])

    def __modelChanged(self, v):
        self.__info_dict['engine'] = v
        # self.setModelInfoByModel()
        self.__db.updateInfo(1, 'engine', v)

    def __streamChecked(self, f):
        self.__stream = f
        self.__db.updateInfo(1, 'stream', f)

    def __finishReasonChecked(self, f):
        self.__finishReason = f


import huggingface_hub
from huggingface_hub import scan_cache_dir

# required tool
# cuda or image with not as good quality
# pip3 install torch torchvision torchaudio --index-url https://download.pytorch.org/whl/cu118
# pip install transformers, accelerate

# want to delete the model cache (which takes a lot of space)
# pip install huggingface_hub["cli"]

commit_hashes = []
cache_dir_result = scan_cache_dir()
for i in cache_dir_result.repos:
    for j in i.revisions:
        print(j.snapshot_path)
        commit_hashes.append(j.commit_hash)
delete_strategy = cache_dir_result.delete_revisions(*commit_hashes)
print("Will free " + delete_strategy.expected_freed_size_str)
delete_strategy.execute()


import subprocess

from qtpy.QtGui import QIcon
from qtpy.QtWidgets import QDialog, QFrame, QPushButton, QHBoxLayout, QVBoxLayout, QWidget, QLabel, \
    QFileDialog, QFormLayout
from qtpy.QtCore import Qt, Signal

from qtpy.QtWidgets import QLineEdit, QMenu, QAction
from qtpy.QtGui import QPixmap
from qtpy.QtWidgets import QGraphicsScene, QGraphicsView

from pyqt_openai.circleProfileImage import RoundedImage


class SingleImageGraphicsView(QGraphicsView):
    def __init__(self):
        super().__init__()
        self.__aspectRatioMode = Qt.KeepAspectRatio
        self.__initVal()

    def __initVal(self):
        self._scene = QGraphicsScene()
        self._p = QPixmap()
        self._item = ''

    def setFilename(self, filename: str):
        self._p = QPixmap(filename)
        self._setPixmap(self._p)

    def setPixmap(self, p):
        self._setPixmap(p)

    def _setPixmap(self, p):
        self._p = p
        self._scene = QGraphicsScene()
        self._item = self._scene.addPixmap(self._p)
        self.setScene(self._scene)
        self.fitInView(self._item, self.__aspectRatioMode)

    def setAspectRatioMode(self, mode):
        self.__aspectRatioMode = mode

    def resizeEvent(self, e):
        if self._item:
            self.fitInView(self._item, self.__aspectRatioMode)
        return super().resizeEvent(e)


class FindPathLineEdit(QLineEdit):
    def __init__(self):
        super().__init__()
        self.__initUi()

    def __initUi(self):
        self.setMouseTracking(True)
        self.setReadOnly(True)
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self.__prepareMenu)

    def mouseMoveEvent(self, e):
        self.__showToolTip()
        return super().mouseMoveEvent(e)

    def __showToolTip(self):
        text = self.text()
        text_width = self.fontMetrics().boundingRect(text).width()

        if text_width > self.width():
            self.setToolTip(text)
        else:
            self.setToolTip('')

    def __prepareMenu(self, pos):
        menu = QMenu(self)
        openDirAction = QAction('Open Path')
        openDirAction.setEnabled(self.text().strip() != '')
        openDirAction.triggered.connect(self.__openPath)
        menu.addAction(openDirAction)
        menu.exec(self.mapToGlobal(pos))

    def __openPath(self):
        filename = self.text()
        path = filename.replace('/', '\\')
        subprocess.Popen(r'explorer /select,"' + path + '"')


class FindPathWidget(QWidget):
    findClicked = Signal()
    added = Signal(str)

    def __init__(self, default_filename: str = ''):
        super().__init__()
        self.__initVal()
        self.__initUi(default_filename)

    def __initVal(self):
        self.__ext_of_files = ''
        self.__directory = False

    def __initUi(self, default_filename: str = ''):
        self.__pathLineEdit = FindPathLineEdit()
        if default_filename:
            self.__pathLineEdit.setText(default_filename)

        self.__pathFindBtn = QPushButton('Find...')

        self.__pathFindBtn.clicked.connect(self.__find)

        self.__pathLineEdit.setMaximumHeight(self.__pathFindBtn.sizeHint().height())

        lay = QHBoxLayout()
        lay.addWidget(self.__pathLineEdit)
        lay.addWidget(self.__pathFindBtn)
        lay.setContentsMargins(0, 0, 0, 0)

        self.setLayout(lay)

    def setLabel(self, text):
        self.layout().insertWidget(0, QLabel(text))

    def setExtOfFiles(self, ext_of_files):
        self.__ext_of_files = ext_of_files

    def getLineEdit(self):
        return self.__pathLineEdit

    def getButton(self):
        return self.__pathFindBtn

    def getFileName(self):
        return self.__pathLineEdit.text()

    def setCustomFind(self, f: bool):
        if f:
            self.__pathFindBtn.clicked.disconnect(self.__find)
            self.__pathFindBtn.clicked.connect(self.__customFind)

    def __customFind(self):
        self.findClicked.emit()

    def __find(self):
        if self.isForDirectory():
            dirname = QFileDialog.getExistingDirectory(None, 'Open Directory', '', QFileDialog.ShowDirsOnly)
            if dirname:
                self.__pathLineEdit.setText(dirname)
                self.added.emit(dirname)
        else:
            str_exp_files_to_open = self.__ext_of_files if self.__ext_of_files else 'All Files (*.*)'
            filename = QFileDialog.getOpenFileName(self, 'Find', '', str_exp_files_to_open)
            if filename[0]:
                filename = filename[0]
                self.__pathLineEdit.setText(filename)
                self.added.emit(filename)

    def setAsDirectory(self, f: bool):
        self.__directory = f

    def isForDirectory(self) -> bool:
        return self.__directory

class CustomizeDialog(QDialog):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__initVal()
        self.__initUi()

    def __initVal(self):
        pass

    def __initUi(self):
        self.setWindowTitle('Customize (working)')
        self.setWindowFlags(Qt.Window | Qt.WindowCloseButtonHint)

        homePageGraphicsView = SingleImageGraphicsView()

        userImage = RoundedImage()
        userImage.setMaximumSize(24, 24)
        userImage.setImage('ico/user.svg')
        AIImage = RoundedImage()
        AIImage.setImage('ico/openai.svg')
        AIImage.setMaximumSize(24, 24)

        findPathWidget1 = FindPathWidget()
        findPathWidget1.added.connect(homePageGraphicsView.setFilename)
        findPathWidget2 = FindPathWidget()
        findPathWidget2.added.connect(userImage.setImage)
        findPathWidget3 = FindPathWidget()
        findPathWidget3.added.connect(AIImage.setImage)

        lay1 = QVBoxLayout()
        lay1.setContentsMargins(0, 0, 0, 0)
        lay1.addWidget(homePageGraphicsView)
        lay1.addWidget(findPathWidget1)
        homePageWidget = QWidget()
        homePageWidget.setLayout(lay1)

        lay2 = QHBoxLayout()
        lay2.setContentsMargins(0, 0, 0, 0)
        lay2.addWidget(userImage)
        lay2.addWidget(findPathWidget2)
        userWidget = QWidget()
        userWidget.setLayout(lay2)

        lay3 = QHBoxLayout()
        lay3.setContentsMargins(0, 0, 0, 0)
        lay3.addWidget(AIImage)
        lay3.addWidget(findPathWidget3)
        aiWidget = QWidget()
        aiWidget.setLayout(lay3)

        lay = QFormLayout()
        lay.addRow('Home Image', homePageWidget)
        lay.addRow('User Image', userWidget)
        lay.addRow('AI Image', aiWidget)

        self.__topWidget = QWidget()
        self.__topWidget.setLayout(lay)

        sep = QFrame()
        sep.setFrameShape(QFrame.HLine)
        sep.setFrameShadow(QFrame.Sunken)

        self.__okBtn = QPushButton('OK')
        self.__okBtn.clicked.connect(self.accept)

        cancelBtn = QPushButton('Cancel')
        cancelBtn.clicked.connect(self.close)

        lay = QHBoxLayout()
        lay.addWidget(self.__okBtn)
        lay.addWidget(cancelBtn)
        lay.setAlignment(Qt.AlignRight)
        lay.setContentsMargins(0, 0, 0, 0)

        okCancelWidget = QWidget()
        okCancelWidget.setLayout(lay)

        lay = QVBoxLayout()
        lay.addWidget(self.__topWidget)
        lay.addWidget(sep)
        lay.addWidget(okCancelWidget)

        self.setLayout(lay)


from qtpy.QtGui import QFont, QIcon
from qtpy.QtWidgets import QWidget, QLabel, QHBoxLayout, QVBoxLayout, QDesktopWidget, QSizePolicy, qApp, QPushButton, \
    QApplication
from qtpy.QtCore import Qt, QPoint, Signal, QTimer, QPropertyAnimation


class NotifierWidget(QWidget):
    doubleClicked = Signal()

    def __init__(self, informative_text='', detailed_text=''):
        super().__init__()
        self.__timerVal = 5000
        self.__initUi(informative_text, detailed_text)

    def __initUi(self, informative_text='', detailed_text=''):
        self.setFixedSize(250, 150)
        self.setWindowFlags(Qt.FramelessWindowHint | Qt.WindowStaysOnTopHint | Qt.SubWindow)

        self.__informativeTextLabel = QLabel(informative_text) if informative_text else QLabel('Informative')
        self.__informativeTextLabel.setFont(QFont('Arial', 15, QFont.Bold))
        self.__detailedTextLabel = QLabel(detailed_text) if detailed_text else QLabel('Detailed')

        closeBtn = QPushButton()
        closeBtn.clicked.connect(self.close)
        closeBtn.setIcon(QIcon('ico/close.svg'))

        lay = QHBoxLayout()
        lay.setContentsMargins(0, 0, 0, 0)

        self.__btnWidget = QWidget()
        self.__btnWidget.setLayout(lay)

        lay = QHBoxLayout()
        lay.setAlignment(Qt.AlignTop | Qt.AlignRight)
        lay.addWidget(closeBtn)
        lay.setContentsMargins(0, 0, 0, 0)

        customMenuBar = QWidget()
        customMenuBar.setLayout(lay)

        lay = QVBoxLayout()
        lay.addWidget(customMenuBar)
        lay.addWidget(self.__informativeTextLabel)
        lay.addWidget(self.__detailedTextLabel)
        lay.addWidget(self.__btnWidget)
        lay.setContentsMargins(0, 0, 0, 0)

        ag = QDesktopWidget().availableGeometry()

        geo = self.geometry()
        geo.moveBottomRight(QPoint(ag.width(), ag.height()))
        self.setGeometry(geo)

        lay.setContentsMargins(8, 8, 8, 8)
        self.setLayout(lay)

    def keyPressEvent(self, e):
        if e.key() == Qt.Key_Escape:
            self.close()

        return super().keyPressEvent(e)

    def addWidgets(self, widgets: list):
        for widget in widgets:
            widget.setSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.MinimumExpanding)
            self.__btnWidget.layout().addWidget(widget)

    def show(self) -> None:
        super().show()
        QApplication.beep()
        self.__timer = QTimer(self)
        self.__timer.timeout.connect(self.__checkTimer)
        self.__timer.start(1000)

    def __checkTimer(self):
        self.__timerVal -= 1000
        if self.__timerVal == 1000:
            self.__showAnimation()
        elif self.__timerVal <= 0:
            self.close()

    def __showAnimation(self):
        self.__animation = QPropertyAnimation(self, b"windowOpacity")
        self.__animation.finished.connect(self.close)
        self.__animation.setDuration(1000)
        self.__animation.setStartValue(1.0)
        self.__animation.setEndValue(0.0)
        self.__animation.start()

    def mouseDoubleClickEvent(self, e):
        self.doubleClicked.emit()
        self.close()
        return super().mouseDoubleClickEvent(e)


# import os
#
# import torch
# from diffusers import StableDiffusionPipeline
#
# model_id = "CompVis/stable-diffusion-v1-4"
# device = "cuda"
#
# pipe = StableDiffusionPipeline.from_pretrained(model_id, torch_dtype=torch.float16)
# pipe = pipe.to(device)
# # save some memory in exchange for a small speed decrease
# # pipe.enable_attention_slicing()
#
# prompt = "bella twins"
# image = pipe(prompt).images[0]
#
# image.save("bella_twins.png")

from diffusers import StableDiffusionPipeline
import torch

model_id = "runwayml/stable-diffusion-v1-5"
pipe = StableDiffusionPipeline.from_pretrained(model_id, torch_dtype=torch.float16, safety_checker=None).to("cuda")
prompt = "Emma Watson with blond hair, crying, white background"

image = pipe(prompt).images[0]
image.save("character.png")


from qtpy.QtCore import Qt
from qtpy.QtGui import QPixmap
from qtpy.QtWidgets import QGraphicsScene, QGraphicsView, QPushButton


class CurrentImageView(QGraphicsView):
    def __init__(self):
        super().__init__()
        self.__aspectRatioMode = Qt.KeepAspectRatio
        self.__gradient_enabled = False
        self.__initVal()
        self.__initUi()

    def __initVal(self):
        self._scene = QGraphicsScene()
        self._p = QPixmap()
        self._item = ''

    def __initUi(self):
        # Create a button
        self.button = QPushButton("Click me", self)
        self.button.hide()
        self.setMouseTracking(True)

    def setFilename(self, filename: str):
        self._p = QPixmap(filename)
        self._scene = QGraphicsScene()
        self._item = self._scene.addPixmap(self._p)

        self.setScene(self._scene)
        self.fitInView(self._item, self.__aspectRatioMode)

    def setAspectRatioMode(self, mode):
        self.__aspectRatioMode = mode

    def enterEvent(self, e):
        # Show the button when the mouse enters the view
        self.button.move(self.rect().x(), self.rect().y())
        self.button.show()
        return super().enterEvent(e)

    def leaveEvent(self, e):
        # Hide the button when the mouse leaves the view
        self.button.hide()
        return super().leaveEvent(e)

    def resizeEvent(self, e):
        if self._item:
            self.fitInView(self.sceneRect(), self.__aspectRatioMode)
        return super().resizeEvent(e)


from datetime import datetime

from qtpy.QtGui import QFont
from qtpy.QtCore import Qt, Signal
from qtpy.QtWidgets import QListWidget, QDialog, QListWidgetItem, QLabel, QHBoxLayout, QWidget, QApplication, QVBoxLayout

from pyqt_openai.inputDialog import InputDialog
from pyqt_openai.svgButton import SvgButton


class ImageItemWidget(QWidget):
    btnClicked = Signal(QListWidgetItem)
    imageUpdated = Signal(int, str)

    def __init__(self, text: str, item: QListWidgetItem, id):
        super().__init__()
        self.__item = item
        self.__id = id
        self.__initUi(text)

    def __initUi(self, text):
        self.__topicLbl = QLabel(text)

        lay = QVBoxLayout()
        lay.addWidget(self.__topicLbl)
        lay.setContentsMargins(0, 0, 0, 0)

        leftWidget = QWidget()
        leftWidget.setLayout(lay)

        editButton = SvgButton()
        editButton.setIcon('ico/edit.svg')
        editButton.setToolTip('Rename')
        editButton.clicked.connect(self.__btnClicked)

        lay = QHBoxLayout()
        lay.setContentsMargins(0, 0, 0, 0)
        lay.addWidget(editButton)
        self.__btnWidget = QWidget()
        self.__btnWidget.setLayout(lay)
        self.__btnWidget.setVisible(False)

        lay = QVBoxLayout()
        lay.addWidget(self.__btnWidget)
        lay.setAlignment(Qt.AlignCenter | Qt.AlignRight)
        lay.setContentsMargins(0, 0, 0, 0)

        rightWidget = QWidget()
        rightWidget.setLayout(lay)

        lay = QHBoxLayout()
        lay.addWidget(leftWidget)
        lay.addWidget(rightWidget)

        self.setLayout(lay)

        # Qt bug - not following by app's font size, update the font on its own after setItemWidget being called
        # to avoid BUG (if there is nothing in qt bug report, i'll report by myself)
        # i don't know this was fixed or not in PySide6 so i will check it out FIXME
        self.__topicLbl.setFont(QApplication.font())

        self.setAutoFillBackground(True)

    def text(self):
        return self.__topicLbl.text()

    def enterEvent(self, e):
        self.__btnWidget.setVisible(True)
        return super().enterEvent(e)

    def leaveEvent(self, e):
        self.__btnWidget.setVisible(False)
        return super().leaveEvent(e)

    def __btnClicked(self):
        dialog = InputDialog('Rename', self.__topicLbl.text(), self)
        reply = dialog.exec()
        if reply == QDialog.Accepted:
            text = dialog.getText()
            self.__topicLbl.setText(text)
            self.imageUpdated.emit(self.__id, text)


class ImageListWidget(QListWidget):
    changed = Signal(QListWidgetItem)
    imageUpdated = Signal(int, str)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__initUi()

    def __initUi(self):
        self.itemClicked.connect(self.__clicked)
        self.currentItemChanged.connect(self.changed)

    def addImage(self, text: str, id: int):
        item = QListWidgetItem()
        item.setFlags(item.flags() | Qt.ItemIsUserCheckable)
        item.setCheckState(Qt.Unchecked)
        widget = ImageItemWidget(text, item, id)
        widget.imageUpdated.connect(self.imageUpdated)
        item.setSizeHint(widget.sizeHint())
        item.setData(Qt.UserRole, id)
        self.insertItem(0, item)
        self.setItemWidget(item, widget)

    def __clicked(self, item):
        potentialChkBoxWidgetInItem = QApplication.widgetAt(self.cursor().pos())
        if isinstance(potentialChkBoxWidgetInItem, QWidget) and potentialChkBoxWidgetInItem.children():
            if isinstance(potentialChkBoxWidgetInItem.children()[0], ImageItemWidget):
                if item.listWidget().itemWidget(item) != None:
                    if item.checkState() == Qt.Checked:
                        item.setCheckState(Qt.Unchecked)
                    else:
                        item.setCheckState(Qt.Checked)

    def toggleState(self, state):
        for i in range(self.count()):
            item = self.item(i)
            if item.checkState() != state:
                item.setCheckState(state)

    def getCheckedRowsIds(self):
        return self.__getFlagRows(Qt.Checked, is_id=True)

    def getUncheckedRowsIds(self):
        return self.__getFlagRows(Qt.Unchecked, is_id=True)

    def __getFlagRows(self, flag: Qt.CheckState, is_id: bool = False):
        flag_lst = []
        for i in range(self.count()):
            item = self.item(i)
            if item.checkState() == flag:
                token = item.data(Qt.UserRole) if is_id else i
                flag_lst.append(token)

        return flag_lst

    def removeCheckedRows(self):
        self.__removeFlagRows(Qt.Checked)

    def removeUncheckedRows(self):
        self.__removeFlagRows(Qt.Unchecked)

    def __removeFlagRows(self, flag):
        flag_lst = self.__getFlagRows(flag)
        flag_lst = reversed(flag_lst)
        for i in flag_lst:
            self.takeItem(i)


import os

from qtpy.QtGui import QPainter
from qtpy.QtSvg import QSvgRenderer
from qtpy.QtWidgets import QLabel


class SvgLabel(QLabel):
    def __init__(self):
        super().__init__()
        self.__renderer = ''

    def paintEvent(self, e):
        painter = QPainter(self)
        painter.restore()
        if self.__renderer:
            self.__renderer.render(painter)
        return super().paintEvent(e)

    def setSvgFile(self, filename: str):
        filename = os.path.join(os.path.dirname(__file__), filename)
        self.__renderer = QSvgRenderer(filename)
        self.resize(self.__renderer.defaultSize())
        length = max(self.sizeHint().width(), self.sizeHint().height())
        self.setFixedSize(length, length)


import requests

from qtpy.QtGui import QFont, QPixmap
from qtpy.QtWidgets import QWidget, QLabel, QGridLayout, QStackedWidget, QSplitter
from qtpy.QtCore import Qt

from pyqt_openai.image_gen_widget.explorerWidget import ExplorerWidget
from pyqt_openai.image_gen_widget.thumbnailView import ThumbnailView


class ViewWidget(QWidget):
    def __init__(self):
        super().__init__()
        self.__initUi()

    def __initUi(self):
        self.__currentImageView = ThumbnailView()
        self.__explorerWidget = ExplorerWidget()
        self.__explorerWidget.clicked.connect(self.__currentImageView.setPixmap)

        imageWidget = QSplitter()
        imageWidget.setOrientation(Qt.Vertical)
        imageWidget.addWidget(self.__currentImageView)
        imageWidget.addWidget(self.__explorerWidget)
        imageWidget.setSizes([700, 300])
        imageWidget.setChildrenCollapsible(False)
        imageWidget.setHandleWidth(2)
        imageWidget.setStyleSheet(
            '''
            QSplitter::handle:vertical
            {
                background: #CCC;
                height: 1px;
            }
            ''')

        homeWidget = QLabel('Home')
        homeWidget.setAlignment(Qt.AlignCenter)
        homeWidget.setFont(QFont('Arial', 32))

        self.__mainWidget = QStackedWidget()
        # self.__mainWidget.addWidget(homeWidget)
        self.__mainWidget.addWidget(imageWidget)

        lay = QGridLayout()
        lay.addWidget(self.__mainWidget)
        lay.setContentsMargins(0, 0, 0, 0)
        self.setLayout(lay)

    def showDallEResult(self, url):
        self.__mainWidget.setCurrentIndex(1)
        self.showImage(url, True)

    def showSdResult(self, image_bin):
        self.__mainWidget.setCurrentIndex(1)
        self.showImage(image_bin, False)

    def showImage(self, arg, f: bool):
        """
        f=True means DALL-E
        f=False means SD
        """
        if f:
            arg = self.__currentImageView.setUrl(arg)
        else:
            self.__currentImageView.setContent(arg)
        self.__explorerWidget.addContent(arg)

    def getExplorerWidget(self):
        return self.__explorerWidget


import openai

# static
# https://platform.openai.com/docs/models/model-endpoint-compatibility
ENDPOINT_DICT = {
    '/v1/chat/completions': ['gpt-4', 'gpt-4-0314', 'gpt-4-32k', 'gpt-4-32k-0314', 'gpt-3.5-turbo', 'gpt-3.5-turbo-0301'],
    '/v1/completions': [
        'text-davinci-003', 'text-davinci-002', 'text-curie-001', 'text-babbage-001', 'text-ada-001', 'davinci',
        'curie', 'babbage', 'ada'
    ],
    '/v1/edits': ['text-davinci-edit-001', 'code-davinci-edit-001'],
    '/v1/audio/transcriptions': ['whisper-1'],
    '/v1/audio/translations': ['whisper-1'],
    '/v1/fine-tunes': ['davinci', 'curie', 'babbage', 'ada'],
    '/v1/embeddings': ['text-embedding-ada-002', 'text-search-ada-doc-001'],
    '/vi/moderations': ['text-moderation-stable', 'text-moderation-latest']
}

def getModelEndpoint(model):
    print(model)
    for k, v in ENDPOINT_DICT.items():
        endpoint_group = list(v)
        if model in endpoint_group:
            return k

# legacy
def getLatestModel():
    return ['gpt-3.5-turbo',
            'gpt-3.5-turbo-0301',
            'text-davinci-003',
            'text-davinci-002',
            'code-davinci-002']

# new
def getCompletionModel():
    return [
        'text-davinci-003',
        'text-davinci-002',
        'text-curie-001',
        'text-babbage-001',
        'text-ada-001'
    ]

def getChatModel():
    return [
        'gpt-3.5-turbo',
        'gpt-3.5-turbo-0301',
        'gpt-4',
        'gpt-4-32k',
    ]

# dynamic
class ModelData:
    def __init__(self):
        super().__init__()
        self.__modelList = []

    def setModelData(self):
        """
        this is only used in completion type, so subtract chat model from it
        :return:
        """
        self.__modelList = list(filter(lambda x: x['id'] not in getChatModel(), openai.Model.list()['data']))

    def getModelData(self):
        return self.__modelList

    def getPermissionProperty(self, model_name, property):
        return [model['permission'][0][property] for model in self.__modelList if model['id'] == model_name][0]


from qtpy.QtCore import Qt, Signal
from qtpy.QtGui import QFont, QTextCursor
from qtpy.QtWidgets import QScrollArea, QCompleter, QVBoxLayout, QToolButton, QMenu, QAction, QWidget, QLabel, \
    QHBoxLayout, QTextEdit, QStackedWidget

from pyqt_openai.sqlite import SqliteDatabase
from pyqt_openai.svgToolButton import SvgToolButton


class ChatBrowser(QScrollArea):
    convUnitUpdated = Signal(int, int, str)

    def __init__(self):
        super().__init__()
        self.__initVal()
        self.__initUi()

    def __initVal(self):
        self.__cur_id = 0

    def __initUi(self):
        self.__homeWidget = QLabel('Home')
        self.__homeWidget.setAlignment(Qt.AlignCenter)
        self.__homeWidget.setFont(QFont('Arial', 32))

        lay = QVBoxLayout()
        lay.setAlignment(Qt.AlignTop)
        lay.setSpacing(0)
        lay.setContentsMargins(0, 0, 0, 0)

        self.__chatWidget = QWidget()
        self.__chatWidget.setLayout(lay)

        widget = QStackedWidget()
        widget.addWidget(self.__homeWidget)
        widget.addWidget(self.__chatWidget)
        self.setWidget(widget)
        self.setWidgetResizable(True)

    def getChatWidget(self):
        return self.__chatWidget

    def showLabel(self, text, user_f, stream_f):
        self.showText(text, stream_f, user_f)
        if not stream_f:
            # change user_f type from bool to int to insert in db
            self.convUnitUpdated.emit(self.__cur_id, int(user_f), text)

    def streamFinished(self):
        self.convUnitUpdated.emit(self.__cur_id, 0, self.getLastResponse())

    def showText(self, text, stream_f, user_f):
        if self.widget().currentWidget() == self.__chatWidget:
            pass
        else:
            self.widget().setCurrentIndex(1)
        self.__setLabel(text, stream_f, user_f)

    def __setLabel(self, text, stream_f, user_f):
        chatLbl = QLabel(text)
        chatLbl.setWordWrap(True)
        chatLbl.setTextInteractionFlags(Qt.TextSelectableByMouse)
        if user_f:
            chatLbl.setStyleSheet('QLabel { padding: 1em }')
            chatLbl.setAlignment(Qt.AlignRight)
        else:
            if stream_f:
                lbl = self.getChatWidget().layout().itemAt(self.getChatWidget().layout().count()-1).widget()
                if isinstance(lbl, QLabel) and lbl.alignment() == Qt.AlignLeft:
                    lbl.setText(lbl.text()+text)
                    return
            chatLbl.setStyleSheet('QLabel { background-color: #DDD; padding: 1em }')
            chatLbl.setAlignment(Qt.AlignLeft)
            chatLbl.setOpenExternalLinks(True)
        self.getChatWidget().layout().addWidget(chatLbl)

    def event(self, e):
        if e.type() == 43:
            self.verticalScrollBar().setSliderPosition(self.verticalScrollBar().maximum())
        return super().event(e)

    def getAllText(self):
        all_text_lst = []
        lay = self.getChatWidget().layout()
        if lay:
            for i in range(lay.count()):
                if lay.itemAt(i) and lay.itemAt(i).widget():
                    widget = lay.itemAt(i).widget()
                    if isinstance(widget, QLabel):
                        all_text_lst.append(widget.text())

        return '\n'.join(all_text_lst)

    def getLastResponse(self):
        lay = self.getChatWidget().layout()
        if lay:
            i = lay.count()-1
            if lay.itemAt(i) and lay.itemAt(i).widget():
                widget = lay.itemAt(i).widget()
                if isinstance(widget, QLabel):
                    return widget.text()
        return ''

    def getEveryResponse(self):
        lay = self.getChatWidget().layout()
        if lay:
            text_lst = []
            for i in range(lay.count()):
                item = lay.itemAt(i)
                if item and item.widget():
                    widget = item.widget()
                    if isinstance(widget, QLabel) and i % 2 == 1:
                        text_lst.append(widget.text())
            return '\n'.join(text_lst)
        else:
            return ''

    def clear(self):
        lay = self.getChatWidget().layout()
        if lay:
            for i in range(lay.count()-1, -1, -1):
                item = lay.itemAt(i)
                if item and item.widget():
                    lay.removeWidget(item.widget())
        self.widget().setCurrentIndex(0)

    def isNew(self):
        return self.widget().currentIndex() == 0

    def setCurId(self, id):
        self.__cur_id = id

    def resetChatWidget(self, id):
        self.clear()
        self.setCurId(id)

    def replaceConv(self, id, conv_data):
        self.clear()
        self.setCurId(id)
        self.widget().setCurrentIndex(1)
        for i in range(len(conv_data)):
            self.__setLabel(conv_data[i], False, not bool(i % 2))


class TextEditPrompt(QTextEdit):
    returnPressed = Signal()

    def __init__(self, db: SqliteDatabase):
        super().__init__()
        self.__initVal(db)
        self.__initUi()

    def __initVal(self, db):
        self.__command_f = False
        self.__completer = None
        self.__db = db

    def __initUi(self):
        self.__initPromptCommandAutocomplete()
        self.setStyleSheet('QTextEdit { border: 1px solid #AAA; } ')

    def setPromptCommandAutocompletedEnabled(self, f):
        self.__command_f = f

    def __initPromptCommandAutocomplete(self):
        p_grp = []
        for group in self.__db.selectPropPromptGroup():
            p_grp_attr = [attr for attr in self.__db.selectPropPromptAttribute(group[0])]
            p_grp_value = ''
            for attr_obj in p_grp_attr:
                name = attr_obj[2]
                value = attr_obj[3]
                if value and value.strip():
                    p_grp_value += f'{name}: {value}\n'
            p_grp.append({'name': group[1], 'value': p_grp_value})

        t_grp = [{'name': obj[1], 'value': obj[2]} for obj in self.__db.selectTemplatePrompt()]

        self.__total_grp = p_grp+t_grp

        # only name
        grp_nm_arr = [obj['name'] for obj in self.__total_grp]

        completer = QCompleter(grp_nm_arr)
        completer.activated.connect(self.__activated)
        self.setCompleter(completer)

    def __activated(self, word):
        for item in self.__total_grp:
            if item['name'] == word:
                v = item['value']
                break
        else:
            # Handle the case when 'b' is not found
            v = ''
        self.textCursor().deletePreviousChar()
        self.insertPlainText(v)

    def setCompleter(self, completer):
        if self.__completer:
            self.__completer.activated.disconnect()

        self.__completer = completer
        self.__completer.setWidget(self)
        self.__completer.setCompletionMode(QCompleter.PopupCompletion)

    def keyPressEvent(self, e):
        if self.__command_f:
            if self.__completer and self.__completer.popup().isVisible():
                if e.key() in (Qt.Key_Enter, Qt.Key_Return, Qt.Key_Escape, Qt.Key_Tab, Qt.Key_Backtab):
                    e.ignore()
                    return

            if e.key() == Qt.Key_Slash:  # Activate completer when "/" is pressed
                self.__initPromptCommandAutocomplete()
                cr = self.cursorRect()
                cr.setWidth(self.__completer.popup().sizeHintForColumn(
                    0) + self.__completer.popup().verticalScrollBar().sizeHint().width())
                self.__completer.complete(cr)
            else:
                self.__completer.popup().hide()

        if e.key() == Qt.Key_Return or e.key() == Qt.Key_Enter:
            if e.modifiers() == Qt.ShiftModifier:
                return super().keyPressEvent(e)
            else:
                self.returnPressed.emit()
        else:
            return super().keyPressEvent(e)


class TextEditPropmtGroup(QWidget):
    textChanged = Signal()

    def __init__(self, db: SqliteDatabase):
        super().__init__()
        self.__initVal(db)
        self.__initUi()

    def __initVal(self, db):
        self.__db = db

    def __initUi(self):
        self.__beginningTextEdit = TextEditPrompt(self.__db)
        self.__beginningTextEdit.textChanged.connect(self.textChanged)
        self.__beginningTextEdit.setPlaceholderText('Beginning')

        self.__textEdit = TextEditPrompt(self.__db)
        self.__textEdit.textChanged.connect(self.textChanged)
        self.__textEdit.setPlaceholderText('Write some text...')

        self.__endingTextEdit = TextEditPrompt(self.__db)
        self.__endingTextEdit.textChanged.connect(self.textChanged)
        self.__endingTextEdit.setPlaceholderText('Ending')

        self.__beginningTextEdit.setVisible(False)
        self.__endingTextEdit.setVisible(False)

        self.__textGroup = [self.__beginningTextEdit, self.__textEdit, self.__endingTextEdit]

        lay = QVBoxLayout()
        for w in self.__textGroup:
            lay.addWidget(w)
        lay.setContentsMargins(0, 0, 0, 0)
        lay.setSpacing(0)

        self.setLayout(lay)

    def setCommandEnabled(self, f: bool):
        for w in self.__textGroup:
            w.setPromptCommandAutocompletedEnabled(f)

    def adjustHeight(self) -> int:
        """
        adjust overall height of text edit group based on their contents and return adjusted height
        :return:
        """
        groupHeight = 0
        for w in self.__textGroup:
            document = w.document()
            height = document.size().height()
            overallHeight = int(height+document.documentMargin())
            w.setMaximumHeight(overallHeight)
            groupHeight += overallHeight
        return groupHeight

    def getGroup(self):
        return self.__textGroup

    def getContent(self):
        b = self.__textGroup[0].toPlainText().strip()
        m = self.__textGroup[1].toPlainText().strip()
        e = self.__textGroup[2].toPlainText().strip()

        content = ''
        if b:
            content = b + '\n'
        content += m
        if e:
            content += '\n' + e

        return content


class Prompt(QWidget):
    def __init__(self, db: SqliteDatabase):
        super().__init__()
        self.__initVal(db)
        self.__initUi()

    def __initVal(self, db):
        self.__db = db

    def __initUi(self):
        self.__textEditGroup = TextEditPropmtGroup(self.__db)
        self.__textEditGroup.textChanged.connect(self.updateHeight)

        settingsBtn = SvgToolButton()
        settingsBtn.setIcon('ico/vertical_three_dots.svg')
        settingsBtn.setToolTip('Prompt Settings')

        # Create the menu
        menu = QMenu(self)

        # Create the actions
        beginningAction = QAction("Show Beginning", self)
        beginningAction.setShortcut('Ctrl+B')
        beginningAction.setCheckable(True)
        beginningAction.toggled.connect(self.__showBeginning)

        endingAction = QAction("Show Ending", self)
        endingAction.setShortcut('Ctrl+E')
        endingAction.setCheckable(True)
        endingAction.toggled.connect(self.__showEnding)

        supportPromptCommandAction = QAction('Support Prompt Command', self)
        supportPromptCommandAction.setShortcut('Ctrl+Shift+P')
        supportPromptCommandAction.setCheckable(True)
        supportPromptCommandAction.toggled.connect(self.__supportPromptCommand)

        # Add the actions to the menu
        menu.addAction(beginningAction)
        menu.addAction(endingAction)
        menu.addAction(supportPromptCommandAction)

        # Connect the button to the menu
        settingsBtn.setMenu(menu)
        settingsBtn.setPopupMode(QToolButton.InstantPopup)

        lay = QVBoxLayout()
        lay.addWidget(settingsBtn)
        lay.setContentsMargins(1, 1, 1, 1)
        lay.setAlignment(Qt.AlignBottom)

        rightWidget = QWidget()
        rightWidget.setLayout(lay)

        lay = QHBoxLayout()
        lay.addWidget(self.__textEditGroup)
        lay.addWidget(rightWidget)
        lay.setContentsMargins(0, 0, 0, 0)
        lay.setSpacing(0)
        self.setLayout(lay)
        self.updateHeight()

    def updateHeight(self):
        overallHeight = self.__textEditGroup.adjustHeight()
        self.setMaximumHeight(overallHeight)

    def getTextEdit(self):
        return self.__textEditGroup.getGroup()[1]

    def getContent(self):
        return self.__textEditGroup.getContent()

    def __showBeginning(self, f):
        self.__textEditGroup.getGroup()[0].setVisible(f)

    def __showEnding(self, f):
        self.__textEditGroup.getGroup()[-1].setVisible(f)

    def __supportPromptCommand(self, f):
        self.__textEditGroup.setCommandEnabled(f)





from qtpy.QtWidgets import QWidget, QDialog, QTableWidget, QVBoxLayout, QHBoxLayout, QHeaderView, QTableWidgetItem, QAbstractItemView
from qtpy.QtCore import Signal, Qt

from pyqt_openai.inputDialog import InputDialog
from pyqt_openai.sqlite import SqliteDatabase
from pyqt_openai.svgButton import SvgButton


class TemplatePage(QWidget):
    updated = Signal(str, str)

    def __init__(self, db: SqliteDatabase):
        super().__init__()
        self.__initVal(db)
        self.__initUi()

    def __initVal(self, db):
        self.__db = db
        self.__previousTemplateArr = self.__db.selectTemplatePrompt()

    def __initUi(self):
        self.__addBtn = SvgButton()
        self.__delBtn = SvgButton()

        self.__addBtn.setIcon('ico/add.svg')
        self.__delBtn.setIcon('ico/delete.svg')

        self.__addBtn.clicked.connect(self.__add)
        self.__delBtn.clicked.connect(self.__delete)

        lay = QHBoxLayout()
        lay.addWidget(self.__addBtn)
        lay.addWidget(self.__delBtn)
        lay.setAlignment(Qt.AlignRight)
        lay.setContentsMargins(0, 0, 0, 0)

        topWidget = QWidget()
        topWidget.setLayout(lay)

        # this template has to be connected with db
        # QSqlTableModel
        self.__templateTable = QTableWidget()
        self.__templateTable.setColumnCount(2)
        self.__templateTable.setHorizontalHeaderLabels(['Name', 'Content'])
        self.__templateTable.horizontalHeader().setSectionResizeMode(1, QHeaderView.Stretch)
        self.__templateTable.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.__templateTable.currentItemChanged.connect(self.__rowChanged)

        self.__templateTable.setRowCount(len(self.__previousTemplateArr))

        for i in range(len(self.__previousTemplateArr)):
            id = self.__previousTemplateArr[i][0]
            name = self.__previousTemplateArr[i][1]
            value = self.__previousTemplateArr[i][2]

            content_item = QTableWidgetItem(name)
            content_item.setTextAlignment(Qt.AlignCenter)
            content_item.setData(Qt.UserRole, id)

            self.__templateTable.setItem(i, 0, content_item)
            self.__templateTable.setItem(i, 1, QTableWidgetItem(value))

        self.__templateTable.itemChanged.connect(self.__saveUpdatedTemplate)

        lay = QVBoxLayout()
        lay.addWidget(topWidget)
        lay.addWidget(self.__templateTable)

        self.setLayout(lay)

    def __rowChanged(self, new_item: QTableWidgetItem, old_item: QTableWidgetItem):
        name_item = self.__templateTable.item(new_item.row(), 0)
        name = name_item.text()
        content = self.__templateTable.item(new_item.row(), 1).text() if new_item.column() == 0 else new_item.text()
        self.updated.emit(content, name)

    def __saveUpdatedTemplate(self, item: QTableWidgetItem):
        name_item = self.__templateTable.item(item.row(), 0)
        id = name_item.data(Qt.UserRole)
        name = name_item.text()
        content = self.__templateTable.item(item.row(), 1).text()
        self.__db.updateTemplatePrompt(id, name, content)

    def __add(self):
        dialog = InputDialog('Name', '', self)
        reply = dialog.exec()
        if reply == QDialog.Accepted:
            self.__templateTable.itemChanged.disconnect(self.__saveUpdatedTemplate)

            text = dialog.getText()
            self.__templateTable.setRowCount(self.__templateTable.rowCount()+1)

            item1 = QTableWidgetItem(text)
            item1.setTextAlignment(Qt.AlignCenter)
            self.__templateTable.setItem(self.__templateTable.rowCount()-1, 0, item1)

            item2 = QTableWidgetItem('')
            item2.setTextAlignment(Qt.AlignCenter)
            self.__templateTable.setItem(self.__templateTable.rowCount()-1, 1, item2)

            id = self.__db.insertTemplatePrompt(text)
            item1.setData(Qt.UserRole, id)

            self.__templateTable.itemChanged.connect(self.__saveUpdatedTemplate)

    def __delete(self):
        for i in sorted(set([i.row() for i in self.__templateTable.selectedIndexes()]), reverse=True):
            id = self.__templateTable.item(i, 0).data(Qt.UserRole)
            self.__db.deleteTemplatePrompt(id)
            self.__templateTable.removeRow(i)


import os.path, posixpath

from qtpy.QtGui import QColor, QPalette, qGray
from qtpy.QtWidgets import QGraphicsColorizeEffect, QWidget, qApp, QToolButton


class SvgToolButton(QToolButton):
    def __init__(self, base_widget: QWidget = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__baseWidget = base_widget
        self.__initVal()
        self.__styleInit()

    def __initVal(self):
        # to set size accordance with scale
        sc = qApp.screens()[0]
        sc.logicalDotsPerInchChanged.connect(self.__scaleChanged)
        self.__size = sc.logicalDotsPerInch() // 4
        self.__padding = self.__border_radius = self.__size // 10
        self.__background_color = 'transparent'
        self.__icon = ''
        self.__animation = ''
        self.installEventFilter(self)
        if self.__baseWidget:
            self.__baseWidget.installEventFilter(self)
            self.__initColorByBaseWidget()
        else:
            self.__hover_color = '#DDDDDD'
            self.__pressed_color = '#FFFFFF'
            self.__checked_color = '#CCCCCC'
            self.__text_color = '#AAAAAA'

    def __initColorByBaseWidget(self):
        self.__base_color = self.__baseWidget.palette().color(QPalette.Base)
        self.__hover_color = self.__getHoverColor(self.__base_color)
        self.__pressed_color = self.__getPressedColor(self.__base_color)
        self.__checked_color = self.__getPressedColor(self.__base_color)
        self.__text_color = self.__getButtonTextColor(self.__base_color)

    def __getColorByFactor(self, base_color, factor):
        r, g, b = base_color.red(), base_color.green(), base_color.blue()
        gray = qGray(r, g, b)
        if gray > 255 // 2:
            color = base_color.darker(factor)
        else:
            color = base_color.lighter(factor)
        return color

    def __getHoverColor(self, base_color):
        hover_factor = 120
        hover_color = self.__getColorByFactor(base_color, hover_factor)
        return hover_color.name()

    def __getPressedColor(self, base_color):
        pressed_factor = 130
        pressed_color = self.__getColorByFactor(base_color, pressed_factor)
        return pressed_color.name()

    def __getCheckedColor(self, base_color):
        return self.__getPressedColor(base_color)

    def __getButtonTextColor(self, base_color):
        r, g, b = base_color.red() ^ 255, base_color.green() ^ 255, base_color.blue() ^ 255
        if r == g == b:
            text_color = QColor(r, g, b)
        else:
            if qGray(r, g, b) > 255 // 2:
                text_color = QColor(255, 255, 255)
            else:
                text_color = QColor(0, 0, 0)
        return text_color.name()

    def __styleInit(self):
        self.__btn_style = f'''
        QAbstractButton
        {{
        border: 0;
        width: {self.__size};
        height: {self.__size};
        image: url({self.__icon});
        background-color: {self.__background_color};
        border-radius: {self.__border_radius};
        padding: {self.__padding};
        color: {self.__text_color};
        }}
        QAbstractButton:hover
        {{
        background-color: {self.__hover_color};
        }}
        QAbstractButton:pressed
        {{
        background-color: {self.__pressed_color};
        }}
        QAbstractButton:checked
        {{
        background-color: {self.__checked_color};
        border: none;
        }}
        '''

        self.setStyleSheet(self.__btn_style)

    def setIcon(self, icon: str):
        self.__icon = os.path.join(os.path.dirname(__file__), icon).replace(os.sep, posixpath.sep)
        self.__styleInit()

    def eventFilter(self, obj, e):
        if obj == self:
            # to change grayscale when button gets disabled
            # if button get enabled/disabled EnableChange will emit
            # so catch the EnabledChange
            if e.type() == 98:
                # change to enabled state
                effect = QGraphicsColorizeEffect()
                effect.setColor(QColor(255, 255, 255))
                if self.isEnabled():
                    effect.setStrength(0)
                else:
                    effect.setStrength(1)
                    effect.setColor(QColor(150, 150, 150))
                self.setGraphicsEffect(effect)
        if obj == self.__baseWidget:
            # catch the StyleChange event of base widget
            if e.type() == 100:
                # if base widget's background is transparent (#ffffff)
                if self.__baseWidget.palette().color(QPalette.Base).name() == '#ffffff':
                    # then check the parent widget's background
                    if self.__baseWidget.parent():
                        if self.__baseWidget.parent().palette().color(QPalette.Base).name() == '#ffffff':
                            pass
                        else:
                            self.__baseWidget = self.__baseWidget.parent()
                self.__initColorByBaseWidget()
                self.__styleInit()
        return super().eventFilter(obj, e)

    def setPadding(self, padding: int):
        self.__padding = padding
        self.__styleInit()

    def setBorderRadius(self, border_radius: int):
        self.__border_radius = border_radius
        self.__styleInit()

    def setBackground(self, background=None):
        if background:
            self.__background_color = background
        else:
            self.__background_color = self.__base_color.name()
        self.__styleInit()

    def setAsCircle(self):
        self.setBorderRadius(self.height() // 2)
        self.__styleInit()

    # to set size accordance with scale
    def __scaleChanged(self, dpi):
        self.__size = dpi // 4
        self.__styleInit()


import json

from qtpy.QtCore import Signal
from qtpy.QtWidgets import QWidget, QCheckBox, QListWidget, QVBoxLayout, QHBoxLayout, QSpacerItem, QSizePolicy, QListWidgetItem, \
    QLabel

from pyqt_openai.convListWidget import ConvListWidget
from pyqt_openai.searchBar import SearchBar
from pyqt_openai.svgButton import SvgButton


class LeftSideBar(QWidget):
    added = Signal()
    changed = Signal(QListWidgetItem)
    deleted = Signal(list)
    convUpdated = Signal(int, str)
    export = Signal(list)

    def __init__(self):
        super().__init__()
        self.__initUi()

    def __initUi(self):
        searchBar = SearchBar()
        searchBar.searched.connect(self.__search)
        searchBar.setPlaceHolder('Search the Conversation...')

        self.__addBtn = SvgButton()
        self.__delBtn = SvgButton()
        self.__saveBtn = SvgButton()

        self.__addBtn.setIcon('ico/add.svg')
        self.__delBtn.setIcon('ico/delete.svg')
        self.__saveBtn.setIcon('ico/save.svg')

        self.__addBtn.setToolTip('Add')
        self.__delBtn.setToolTip('Delete')
        self.__saveBtn.setToolTip('Save')

        self.__addBtn.clicked.connect(self.__addClicked)
        self.__delBtn.clicked.connect(self.__deleteClicked)
        self.__saveBtn.clicked.connect(self.__saveClicked)

        self.__allCheckBox = QCheckBox('Check All')
        self.__allCheckBox.stateChanged.connect(self.__stateChanged)

        lay = QHBoxLayout()
        lay.addWidget(self.__allCheckBox)
        lay.addSpacerItem(QSpacerItem(10, 10, QSizePolicy.MinimumExpanding))
        lay.addWidget(self.__addBtn)
        lay.addWidget(self.__delBtn)
        lay.addWidget(self.__saveBtn)
        lay.setContentsMargins(0, 0, 0, 0)

        navWidget = QWidget()
        navWidget.setLayout(lay)

        lay = QVBoxLayout()
        lay.addWidget(navWidget)
        lay.addWidget(searchBar)

        topWidget = QWidget()
        topWidget.setLayout(lay)
        lay.setContentsMargins(0, 0, 0, 0)

        self.__convListWidget = ConvListWidget()
        self.__convListWidget.changed.connect(self.changed)
        self.__convListWidget.convUpdated.connect(self.convUpdated)

        lay = QVBoxLayout()
        lay.addWidget(topWidget)
        lay.addWidget(self.__convListWidget)

        self.setLayout(lay)

    def __addClicked(self):
        self.added.emit()

    def addToList(self, id):
        self.__convListWidget.addConv('New Chat', id)
        self.__convListWidget.setCurrentRow(0)

    def isCurrentConvExists(self):
        return self.__convListWidget.count() > 0 and self.__convListWidget.currentItem()

    def __deleteClicked(self):
        # get the ID of row, not actual index (because list is in a stacked form)
        rows = self.__convListWidget.getCheckedRowsIds()
        self.__convListWidget.removeCheckedRows()
        self.deleted.emit(rows)
        self.__allCheckBox.setChecked(False)

    def __saveClicked(self):
        self.export.emit(self.__convListWidget.getUncheckedRowsIds())

    def __stateChanged(self, f):
        self.__convListWidget.toggleState(f)

    def __search(self, text):
        for i in range(self.__convListWidget.count()):
            item = self.__convListWidget.item(i)
            widget = self.__convListWidget.itemWidget(item)
            item.setHidden(False if text.lower() in widget.text().lower() else True)

    def initHistory(self, db):
        try:
            conv_lst = db.selectAllConv()
            for conv in conv_lst:
                id, title = conv[0], conv[1]
                self.__convListWidget.addConv(title, id)
        except Exception as e:
            print(e)


import sqlite3, json, shutil


class SqliteDatabase:
    """
    functions which only meant to be used frequently are defined.

    if there is no functions you want to use, use ``getCursor`` instead
    """
    def __init__(self):
        super().__init__()
        self.__initVal()
        self.__initDb()
        self.__createConv()

    def __initVal(self):
        # db names
        self.__db_filename = 'conv.db'

        # conv table names
        self.__conv_tb_nm = 'conv_tb'
        self.__conv_tb_tr_nm = 'conv_tr'
        self.__conv_unit_tb_nm = 'conv_unit_tb'

        # info table names
        self.__info_tb_nm = 'info_tb'
        self.__completion_info_tb_nm = 'info_completion_tb'

        # prompt table
        self.__prop_prompt_group_tb_nm = 'prop_prompt_grp_tb'
        self.__prop_prompt_unit_tb_nm = 'prop_prompt_unit_tb'

        self.__template_prompt_tb_nm = 'template_prompt_tb'

        # model type (chat, etc.)
        self.__model_type = 1

        # default value of each properties based on https://platform.openai.com/docs/api-reference/chat/create
        # GPT-3.5(ChatGPT), GPT-4
        self.__chat_default_value = {
            'engine': "gpt-3.5-turbo",
            'system': "You are a helpful assistant.",
            'temperature': 0.7,
            # -1 means infinite, not currently used in this application
            'max_tokens': -1,
            'top_p': 1,
            'frequency_penalty': 0,
            'presence_penalty': 0,
            'stream': True
        }

        # GPT-3, etc.
        self.__completion_default_value = {
            'engine': "text-davinci-003",
            'temperature': 0.7,
            'max_tokens': 4096,
            'top_p': 1,
            'frequency_penalty': 0,
            'presence_penalty': 0,
        }

        # DALL-E
        self.__image_default_value = {
            'engine': "DALL-E",
            'n': 1,
            'width': 1024,
            'height': 1024,
            # 'response_format':
        }

        self.__each_info_dict = {1: [self.__info_tb_nm, self.__chat_default_value],
                                 2: [self.__completion_info_tb_nm, self.__completion_default_value], }

        self.__prop_prompt_unit_default_value = [{'name': 'Task', 'text': ''},
                                                 {'name': 'Topic', 'text': ''},
                                                 {'name': 'Style', 'text': ''},
                                                 {'name': 'Tone', 'text': ''},
                                                 {'name': 'Audience', 'text': ''},
                                                 {'name': 'Length', 'text': ''},
                                                 {'name': 'Form', 'text': ''}]

        # based on Alex Brogan's prompt example
        self.__template_prompt_default_value = [
            {'name': 'Sample 1',
             'text': 'Identify the 20% of [topic or skill] that will yield 80% of the desired results and provide a focused learning plan to master it.'},
             {'name': 'Sample 2',
              'text': 'Explain [topic or skill] in the simplest terms possible as if teaching it to a complete beginner. Identify gaps in my understanding and suggest resources to fill them.'},
              {'name': 'Sample 3',
               'text': 'Create a study plan that mixes different topics or skills within [subject area] to help me develop a more robust understanding and facilitate connections between them.'},
               {'name': 'Sample 4',
                'text': 'Design a spaced repetition schedule for me to effectively review [topic or skill] over time, ensuring better retention and recall.'},
                {'name': 'Sample 5',
                 'text': 'Help me create mental models or analogies to better understand and remember key concepts in [topic or skill].'},
                 {'name': 'Sample 6',
                  'text': 'Suggest various learning resources (e.g., videos, books, podcasts, interactive exercises) for [topic or skill] that cater to different learning styles.'},
                  {'name': 'Sample 7',
                   'text': 'Provide me with a series of challenging questions or problems related to [topic or skill] to test my understanding and improve long-term retention.'},
                   {'name': 'Sample 8',
                    'text': 'Transform key concepts or lessons from [topic or skill] into engaging stories or narratives to help me better remember and understand the material.'},
                    {'name': 'Sample 9',
                     'text': 'Design a deliberate practice routine for [topic or skill], focusing on my weaknesses and providing regular feedback for improvement.'},
                     {'name': 'Sample 10',
                      'text': 'Guide me through a visualization exercise to help me internalize [topic or skill] and imagine myself succesfully applying it in real-life situations.'}
        ]

    def __initDb(self):
        try:
            # Connect to the database (create a new file if it doesn't exist)
            self.__conn = sqlite3.connect(self.__db_filename)
            self.__conn.execute('PRAGMA foreign_keys = ON;')
            self.__conn.commit()

            self.__c = self.__conn.cursor()
            self.__createInfo()
        except sqlite3.Error as e:
            print(f"An error occurred while connecting to the database: {e}")
            raise

    def __createChat(self):
        # Check if the table exists
        self.__c.execute(f"SELECT count(*) FROM sqlite_master WHERE type='table' AND name='{self.__info_tb_nm}'")
        if self.__c.fetchone()[0] == 1:
            # Check if each column already exists in the table
            existing_columns = [row[1] for row in self.__c.execute(f"PRAGMA table_info({self.__info_tb_nm});")]
            new_columns = [col for col in self.__chat_default_value.keys() if col not in existing_columns]
            # Add the new columns if they don't already exist
            # TODO specify the type
            for col in new_columns:
                d_value = self.__chat_default_value[col]
                if isinstance(self.__chat_default_value[col], str):
                    d_value = f'"{d_value}"' if len(self.__chat_default_value[col].split()) > 0 else d_value
                self.__c.execute(f"ALTER TABLE {self.__info_tb_nm} ADD COLUMN {col} DEFAULT {d_value}")
        else:
            self.__c.execute(f'''CREATE TABLE {self.__info_tb_nm}
                                     (id INTEGER PRIMARY KEY,
                                      engine VARCHAR(50) DEFAULT '{self.__chat_default_value['engine']}',
                                      system TEXT DEFAULT '{self.__chat_default_value['system']}',
                                      temperature INTEGER DEFAULT {self.__chat_default_value['temperature']},
                                      max_tokens INTEGER DEFAULT {self.__chat_default_value['max_tokens']},
                                      top_p INTEGER DEFAULT {self.__chat_default_value['top_p']},
                                      frequency_penalty INTEGER DEFAULT {self.__chat_default_value['frequency_penalty']},
                                      presence_penalty INTEGER DEFAULT {self.__chat_default_value['presence_penalty']},
                                      stream BOOL DEFAULT {self.__chat_default_value['stream']},

                                      update_dt DATETIME DEFAULT CURRENT_TIMESTAMP,
                                      insert_dt DATETIME DEFAULT CURRENT_TIMESTAMP)''')

            # Commit the transaction
            self.__conn.commit()

            # insert default record
            self.__c.execute(f'''INSERT INTO {self.__info_tb_nm}
                                            (
                                                engine,
                                                system,
                                                temperature,
                                                max_tokens,
                                                top_p,
                                                frequency_penalty,
                                                presence_penalty,
                                                stream
                                            ) VALUES
                                            (
                                                {','.join(['?' for _ in range(len(self.__chat_default_value))])}
                                            )
                                         ''', tuple(self.__chat_default_value.values()))

    def __createCompletion(self):
        # Check if the table exists
        self.__c.execute(f"SELECT count(*) FROM sqlite_master WHERE type='table' AND name='{self.__completion_info_tb_nm}'")
        if self.__c.fetchone()[0] == 1:
            pass
        else:
            self.__c.execute(f'''CREATE TABLE {self.__completion_info_tb_nm}
                                             (id INTEGER PRIMARY KEY,
                                              engine VARCHAR(50) DEFAULT '{self.__completion_default_value['engine']}',
                                              temperature INTEGER DEFAULT {self.__completion_default_value['temperature']},
                                              max_tokens INTEGER DEFAULT {self.__completion_default_value['max_tokens']},
                                              top_p INTEGER DEFAULT {self.__completion_default_value['top_p']},
                                              frequency_penalty INTEGER DEFAULT {self.__completion_default_value['frequency_penalty']},
                                              presence_penalty INTEGER DEFAULT {self.__completion_default_value['presence_penalty']},

                                              update_dt DATETIME DEFAULT CURRENT_TIMESTAMP,
                                              insert_dt DATETIME DEFAULT CURRENT_TIMESTAMP)''')

            # Commit the transaction
            self.__conn.commit()

            # insert default record
            self.__c.execute(f'''INSERT INTO {self.__completion_info_tb_nm}
                                                    (
                                                        engine,
                                                        temperature,
                                                        max_tokens,
                                                        top_p,
                                                        frequency_penalty,
                                                        presence_penalty
                                                    ) VALUES
                                                    (
                                                        {','.join(['?' for _ in range(len(self.__completion_default_value))])}
                                                    )
                                                 ''', tuple(self.__completion_default_value.values()))

    def __createPropPromptGroup(self):
        self.__c.execute(
            f"SELECT count(*) FROM sqlite_master WHERE type='table' AND name='{self.__prop_prompt_group_tb_nm}'")
        if self.__c.fetchone()[0] == 1:
            pass
        else:
            self.__c.execute(f'''CREATE TABLE {self.__prop_prompt_group_tb_nm}
                                                 (id INTEGER PRIMARY KEY,
                                                  name VARCHAR(50),
                                                  
                                                  update_dt DATETIME DEFAULT CURRENT_TIMESTAMP,
                                                  insert_dt DATETIME DEFAULT CURRENT_TIMESTAMP)''')

            # Commit the transaction
            self.__conn.commit()

            self.insertPropPromptGroup('Default')

    def createDefaultPropPromptAttributes(self, id_fk):
        self.__c.execute(
            f"SELECT count(*) FROM sqlite_master WHERE type='table' AND name='{self.__prop_prompt_unit_tb_nm}{id_fk}'")
        if self.__c.fetchone()[0] == 1:
            pass
        else:
            self.__c.execute(f'''CREATE TABLE {self.__prop_prompt_unit_tb_nm}{id_fk}
                                                             (id INTEGER PRIMARY KEY,
                                                              id_fk INTEGER,
                                                              name VARCHAR(50),
                                                              text TEXT,
                                                              update_dt DATETIME DEFAULT CURRENT_TIMESTAMP,
                                                              insert_dt DATETIME DEFAULT CURRENT_TIMESTAMP,
                                                              FOREIGN KEY (id_fk) REFERENCES {self.__prop_prompt_group_tb_nm}(id)
                                                              ON DELETE CASCADE)''')

        # insert default property group
        for obj in self.__prop_prompt_unit_default_value:
            lst = [id_fk] + list(tuple(obj.values()))
            self.__c.execute(f"INSERT INTO {self.__prop_prompt_unit_tb_nm}{id_fk} (id_fk, name, text) VALUES (?, ?, ?)", tuple(lst))

        # Commit the transaction
        self.__conn.commit()

    def selectPropPromptAttribute(self, id):
        try:
            self.__c.execute(f'SELECT * FROM {self.__prop_prompt_unit_tb_nm}{id}')
            return self.__c.fetchall()
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def insertPropPromptAttribute(self, id, name):
        try:
            # Insert a row into the table
            self.__c.execute(f'INSERT INTO {self.__prop_prompt_unit_tb_nm}{id} (id_fk, name) VALUES (?, ?)', (id, name,))
            new_id = self.__c.lastrowid
            # Commit the transaction
            self.__conn.commit()
            return new_id
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def updatePropPromptAttribute(self, p_id, id, name, text):
        try:
            self.__c.execute(f'UPDATE {self.__prop_prompt_unit_tb_nm}{p_id} SET name=?, text=? WHERE id={id}', (name, text))
            self.__conn.commit()
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def deletePropPromptAttribute(self, p_id, id):
        try:
            self.__c.execute(f'DELETE FROM {self.__prop_prompt_unit_tb_nm}{p_id} WHERE id={id}')
            self.__conn.commit()
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def selectPropPromptGroup(self):
        try:
            self.__c.execute(f'SELECT * FROM {self.__prop_prompt_group_tb_nm}')
            return self.__c.fetchall()
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def selectPropPromptGroupId(self, id):
        try:
            self.__c.execute(f'SELECT * FROM {self.__prop_prompt_group_tb_nm} WHERE id={id}')
            return self.__c.fetchone()
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def insertPropPromptGroup(self, name):
        try:
            # Insert a row into the table
            self.__c.execute(f'INSERT INTO {self.__prop_prompt_group_tb_nm} (name) VALUES (?)', (name,))
            new_id = self.__c.lastrowid
            # Commit the transaction
            self.__conn.commit()
            # insert default attributes
            self.createDefaultPropPromptAttributes(new_id)
            # TODO abcd make others insert statement return like this
            return new_id
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def updatePropPromptGroup(self, id, name):
        try:
            self.__c.execute(f'UPDATE {self.__prop_prompt_group_tb_nm} SET name=(?) WHERE id={id}', (name,))
            self.__conn.commit()
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def deletePropPromptGroup(self, id):
        try:
            self.__c.execute(f'DELETE FROM {self.__prop_prompt_group_tb_nm} WHERE id={id}')
            self.__conn.commit()
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def __createTemplatePrompt(self):
        self.__c.execute(
            f"SELECT count(*) FROM sqlite_master WHERE type='table' AND name='{self.__template_prompt_tb_nm}'")
        if self.__c.fetchone()[0] == 1:
            pass
        else:
            self.__c.execute(f'''CREATE TABLE {self.__template_prompt_tb_nm}
                                                 (id INTEGER PRIMARY KEY,
                                                  name VARCHAR(50),
                                                  text TEXT,
                                                  update_dt DATETIME DEFAULT CURRENT_TIMESTAMP,
                                                  insert_dt DATETIME DEFAULT CURRENT_TIMESTAMP)''')

            # Commit the transaction
            self.__conn.commit()

            # insert default template set
            for obj in self.__template_prompt_default_value:
                self.__c.execute(f"INSERT INTO {self.__template_prompt_tb_nm} (name, text) VALUES (?, ?)", tuple(obj.values()))

    def selectTemplatePrompt(self):
        try:
            self.__c.execute(f'SELECT * FROM {self.__template_prompt_tb_nm}')
            return self.__c.fetchall()
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def insertTemplatePrompt(self, name):
        try:
            # Insert a row into the table
            self.__c.execute(f"INSERT INTO {self.__template_prompt_tb_nm} (name, text) VALUES (?, ?)",
                             (name, ''))
            new_id = self.__c.lastrowid
            # Commit the transaction
            self.__conn.commit()
            return new_id
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def updateTemplatePrompt(self, id, name, text):
        try:
            self.__c.execute(f'UPDATE {self.__template_prompt_tb_nm} SET name=(?), text=(?) WHERE id={id}', (name, text))
            self.__conn.commit()
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def deleteTemplatePrompt(self, id):
        try:
            self.__c.execute(f'DELETE FROM {self.__template_prompt_tb_nm} WHERE id={id}')
            self.__conn.commit()
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def __createInfo(self):
        try:
            # chat and completion information
            self.__createChat()
            self.__createCompletion()

            # prompt information
            self.__createPropPromptGroup()
            self.__createTemplatePrompt()

            # Commit the transaction
            self.__conn.commit()
        except sqlite3.Error as e:
            print(f"An error occurred while creating the table: {e}")
            raise

    def __createConv(self):
        try:
            # Check if the table exists
            self.__c.execute(f"SELECT count(*) FROM sqlite_master WHERE type='table' AND name='{self.__conv_tb_nm}'")
            if self.__c.fetchone()[0] == 1:
                # each conv table already exists
                pass
            else:
                # Create a table with update_dt and insert_dt columns
                self.__c.execute(f'''CREATE TABLE {self.__conv_tb_nm}
                             (id INTEGER PRIMARY KEY,
                              name TEXT,
                              update_dt DATETIME DEFAULT CURRENT_TIMESTAMP,
                              insert_dt DATETIME DEFAULT CURRENT_TIMESTAMP)''')
                # Create a trigger to update the update_dt column with the current timestamp
                self.__c.execute(f'''CREATE TRIGGER {self.__conv_tb_tr_nm}
                             AFTER UPDATE ON {self.__conv_tb_nm}
                             FOR EACH ROW
                             BEGIN
                               UPDATE {self.__conv_tb_nm}
                               SET update_dt=CURRENT_TIMESTAMP
                               WHERE id=OLD.id;
                             END;''')
                # Commit the transaction
                self.__conn.commit()
        except sqlite3.Error as e:
            print(f"An error occurred while creating the table: {e}")
            raise

    def selectAllInfo(self):
        """
        select all info
        FIXME
        """
        try:
            # filter bool type fields
            bool_type_column = [row[1] for row in self.__c.execute(f'PRAGMA table_info({self.__info_tb_nm})').fetchall() if row[2] == 'BOOL']

            # Execute the SELECT statement
            self.__c.execute(f'SELECT {",".join(list(self.__chat_default_value.keys()))} FROM {self.__info_tb_nm}')

            # Get the column names
            column_names = [description[0] for description in self.__c.description]

            rows = self.__c.fetchall()

            info_dict_arr = []

            # Get the rows with field names and values
            for row in rows:
                info_dict = {}
                for i, value in enumerate(row):
                    if column_names[i] in bool_type_column:
                        value = True if value == 1 else False
                    info_dict[column_names[i]] = value
                info_dict_arr.append(info_dict)

            return info_dict_arr
        except sqlite3.Error as e:
            print(f"An error occurred while creating the table: {e}")
            raise

    def selectInfo(self, id=None):
        """
        select specific info
        default value is 1 (chat - gpt3.5, gpt4, etc.)
        """
        try:
            # default value is 1 (chat - gpt3.5, gpt4, etc.)
            id = id if id else self.__model_type

            # filter bool type fields
            bool_type_column = [row[1] for row in self.__c.execute(f'PRAGMA table_info({self.__each_info_dict[id][0]})').fetchall() if row[2] == 'BOOL']

            # Execute the SELECT statement
            self.__c.execute(f'SELECT {",".join(list(self.__each_info_dict[id][1].keys()))} FROM {self.__each_info_dict[id][0]}')

            # Get the column names
            column_names = [description[0] for description in self.__c.description]

            row = self.__c.fetchone()

            info_dict = {}
            for i, value in enumerate(row):
                if column_names[i] in bool_type_column:
                    value = True if value == 1 else False
                info_dict[column_names[i]] = value

            return info_dict
        except sqlite3.Error as e:
            print(f"An error occurred while creating the table: {e}")
            raise

    def updateInfo(self, id, field, value):
        try:
            self.__c.execute(f'UPDATE {self.__each_info_dict[id][0]} SET {field}=(?) WHERE id=1', (value,))
            self.__conn.commit()
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def selectAllConv(self):
        """
        select all conv
        """
        try:
            self.__c.execute(f'SELECT * FROM {self.__conv_tb_nm}')
            return self.__c.fetchall()
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def selectConv(self, id):
        """
        select specific conv
        """
        try:
            self.__c.execute(f'SELECT * FROM {self.__conv_tb_nm} WHERE id={id}')
            return self.__c.fetchone()
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def insertConv(self, name):
        try:
            # Insert a row into the table
            self.__c.execute(f'INSERT INTO {self.__conv_tb_nm} (name) VALUES (?)', (name,))
            new_id = self.__c.lastrowid
            # Commit the transaction
            self.__conn.commit()
            self.__createConvUnit(new_id)
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def updateConv(self, id, name):
        try:
            self.__c.execute(f'UPDATE {self.__conv_tb_nm} SET name=(?) WHERE id={id}', (name,))
            self.__conn.commit()
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def deleteConv(self, id):
        try:
            self.__c.execute(f'DELETE FROM {self.__conv_tb_nm} WHERE id={id}')
            self.__conn.commit()
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def __createConvUnit(self, id_fk):
        try:
            # Check if the table exists
            self.__c.execute(f"SELECT count(*) FROM sqlite_master WHERE type='table' AND name='{self.__conv_unit_tb_nm}{id_fk}'")
            if self.__c.fetchone()[0] == 1:
                # each conv table already exists
                pass
            else:
                self.__c.execute(f'''CREATE TABLE {self.__conv_unit_tb_nm}{id_fk}
                                         (id INTEGER PRIMARY KEY,
                                          id_fk INTEGER,
                                          is_user INTEGER,
                                          conv TEXT,
                                          update_dt DATETIME DEFAULT CURRENT_TIMESTAMP,
                                          insert_dt DATETIME DEFAULT CURRENT_TIMESTAMP,
                                          FOREIGN KEY (id_fk) REFERENCES {self.__conv_tb_nm}(id) ON DELETE CASCADE)''')

                # insert trigger
                self.__c.execute(f'''
                    CREATE TRIGGER conv_tb_updated_by_unit_inserted_tr{id_fk}
                    AFTER INSERT ON {self.__conv_unit_tb_nm}{id_fk}
                    BEGIN
                      UPDATE {self.__conv_tb_nm} SET update_dt = CURRENT_TIMESTAMP WHERE id = NEW.id_fk;
                    END
                ''')

                # update trigger
                self.__c.execute(f'''
                    CREATE TRIGGER conv_tb_updated_by_unit_updated_tr{id_fk}
                    AFTER UPDATE ON {self.__conv_unit_tb_nm}{id_fk}
                    BEGIN
                      UPDATE {self.__conv_tb_nm} SET update_dt = CURRENT_TIMESTAMP WHERE id = NEW.id_fk;
                    END
                ''')

                # delete trigger
                self.__c.execute(f'''
                    CREATE TRIGGER conv_tb_updated_by_unit_deleted_tr{id_fk}
                    AFTER DELETE ON {self.__conv_unit_tb_nm}{id_fk}
                    BEGIN
                      UPDATE {self.__conv_tb_nm} SET update_dt = CURRENT_TIMESTAMP WHERE id = OLD.id_fk;
                    END
                ''')
                # Commit the transaction
                self.__conn.commit()
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def selectConvUnit(self, id):
        self.__c.execute(f'SELECT * FROM {self.getConvUnitTableName()}{id}')
        return [elem[3] for elem in self.__c.fetchall()]

    def insertConvUnit(self, id, user_f, conv):
        try:
            # Insert a row into the table
            self.__c.execute(
                f'INSERT INTO {self.__conv_unit_tb_nm}{id} (id_fk, is_user, conv) VALUES (?, ?, ?)',
                (id, user_f, conv,))
            # Commit the transaction
            self.__conn.commit()
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def setModelType(self, model_type: int):
        """
        :param model_type: it starts from 1
        :return:
        """
        self.__model_type = model_type

    def export(self, ids, saved_filename):
        shutil.copy2(self.__db_filename, saved_filename)
        conn = sqlite3.connect(saved_filename)

        placeholders = ','.join('?' for _ in ids)
        cursor = conn.cursor()
        for i in range(len(ids)):
            delete_conv_q = f"DELETE FROM {self.__conv_tb_nm} WHERE id in ({placeholders})"
            cursor.execute(delete_conv_q, ids)
            drop_conv_unit_tb_q = f"DROP TABLE {self.__conv_unit_tb_nm}{ids[i]}"
            cursor.execute(drop_conv_unit_tb_q)
            conn.commit()

        conn.close()

    def convertJsonIntoSql(self):
        try:
            # Read data from json
            with open('test/conv_history.json', 'r') as f:
                data = json.load(f)
                for obj in list(data.values())[0]:
                    id, title, conv_data = obj.values()
                    # start from 1 in sql unlike json which starts from 0
                    id += 1
                    self.insertConv(title)
                    for i in range(len(conv_data)):
                        # Insert a row into the table
                        self.__c.execute(
                            f'INSERT INTO {self.__conv_unit_tb_nm}{id} (id_fk, is_user, conv) VALUES (?, ?, ?)',
                            (id, i % 2 == 0, conv_data[i],))
                        # Commit the transaction
                        self.__conn.commit()
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def getCursor(self):
        return self.__c

    def getConvTableName(self):
        return self.__conv_tb_nm

    def getConvUnitTableName(self):
        return self.__conv_unit_tb_nm

    def close(self):
        self.__conn.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # Close the connection
        self.__conn.close()


from qtpy.QtCore import Signal, Qt
from qtpy.QtWidgets import QTableWidget, QSizePolicy, QPushButton, QSpacerItem, QStackedWidget, QLabel, \
    QAbstractItemView, QTableWidgetItem, QHeaderView, QHBoxLayout, \
    QVBoxLayout, QWidget, QDialog, QListWidget, QListWidgetItem, QApplication, QSplitter

from pyqt_openai.inputDialog import InputDialog
from pyqt_openai.sqlite import SqliteDatabase
from pyqt_openai.svgButton import SvgButton


class PropGroupList(QWidget):
    added = Signal(int)
    deleted = Signal(int)
    currentRowChanged = Signal(int)

    def __init__(self, db: SqliteDatabase):
        super().__init__()
        self.__initVal(db)
        self.__initUi()

    def __initVal(self, db):
        self.__db = db

    def __initUi(self):
        self.__addBtn = SvgButton()
        self.__delBtn = SvgButton()

        self.__addBtn.setIcon('ico/add.svg')
        self.__delBtn.setIcon('ico/delete.svg')

        self.__addBtn.clicked.connect(self.__addGroup)
        self.__delBtn.clicked.connect(self.__deleteGroup)

        lay = QHBoxLayout()
        lay.addWidget(QLabel('Property Group'))
        lay.addSpacerItem(QSpacerItem(10, 10, QSizePolicy.MinimumExpanding))
        lay.addWidget(self.__addBtn)
        lay.addWidget(self.__delBtn)
        lay.setAlignment(Qt.AlignRight)
        lay.setContentsMargins(0, 0, 0, 0)

        topWidget = QWidget()
        topWidget.setLayout(lay)

        defaultPropPromptGroupArr = self.__db.selectPropPromptGroup()

        self.__propList = QListWidget()

        # TODO abcd
        for group in defaultPropPromptGroupArr:
            id = group[0]
            name = group[1]
            self.__addGroupItem(id, name)

        self.__propList.currentRowChanged.connect(self.currentRowChanged)

        lay = QVBoxLayout()
        lay.addWidget(topWidget)
        lay.addWidget(self.__propList)
        lay.setContentsMargins(0, 0, 5, 0)

        self.setLayout(lay)

        self.__propList.setCurrentRow(0)

    def __addGroupItem(self, id, name):
        item = QListWidgetItem()
        item.setData(Qt.UserRole, id)
        item.setText(name)
        self.__propList.addItem(item)
        self.__propList.setCurrentItem(item)
        self.added.emit(id)

    def __addGroup(self):
        dialog = InputDialog('Add', '', self)
        reply = dialog.exec()
        if reply == QDialog.Accepted:
            name = dialog.getText()
            id = self.__db.insertPropPromptGroup(name)
            self.__addGroupItem(id, name)

    def __deleteGroup(self):
        i = self.__propList.currentRow()
        item = self.__propList.takeItem(i)
        id = item.data(Qt.UserRole)
        self.__db.deletePropPromptGroup(id)
        self.deleted.emit(i)


class PropTable(QWidget):
    """
    benchmarked https://gptforwork.com/tools/prompt-generator
    """
    updated = Signal(str)

    def __init__(self, db: SqliteDatabase, id):
        super().__init__()
        self.__initVal(db, id)
        self.__initUi()

    def __initVal(self, db, id):
        self.__db = db
        self.__id = id

        self.__title = self.__db.selectPropPromptGroupId(self.__id)[1]
        self.__previousPromptPropArr = self.__db.selectPropPromptAttribute(self.__id)

    def __initUi(self):
        self.__addBtn = SvgButton()
        self.__delBtn = SvgButton()

        self.__addBtn.setIcon('ico/add.svg')
        self.__delBtn.setIcon('ico/delete.svg')

        self.__addBtn.clicked.connect(self.__add)
        self.__delBtn.clicked.connect(self.__delete)

        lay = QHBoxLayout()
        lay.addWidget(QLabel(self.__title))
        lay.addSpacerItem(QSpacerItem(10, 10, QSizePolicy.MinimumExpanding))
        lay.addWidget(self.__addBtn)
        lay.addWidget(self.__delBtn)
        lay.setAlignment(Qt.AlignRight)
        lay.setContentsMargins(0, 0, 0, 0)

        topWidget = QWidget()
        topWidget.setLayout(lay)

        self.__table = QTableWidget()
        self.__table.setColumnCount(2)
        self.__table.setRowCount(len(self.__previousPromptPropArr))
        self.__table.horizontalHeader().setSectionResizeMode(1, QHeaderView.Stretch)
        self.__table.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.__table.setHorizontalHeaderLabels(['Name', 'Value'])

        for i in range(len(self.__previousPromptPropArr)):
            name = self.__previousPromptPropArr[i][2]
            value = self.__previousPromptPropArr[i][3]

            item1 = QTableWidgetItem(name)
            item1.setData(Qt.UserRole, self.__previousPromptPropArr[i][0])
            item1.setTextAlignment(Qt.AlignCenter)

            item2 = QTableWidgetItem(value)
            item2.setTextAlignment(Qt.AlignCenter)

            self.__table.setItem(i, 0, item1)
            self.__table.setItem(i, 1, item2)

        self.__table.itemChanged.connect(self.__generatePropPrompt)
        self.__table.itemChanged.connect(self.__saveChangedPropPrompt)

        lay = QVBoxLayout()
        lay.addWidget(topWidget)
        lay.addWidget(self.__table)
        lay.setContentsMargins(5, 0, 0, 0)

        self.setLayout(lay)

    def getPromptText(self):
        prompt_text = ''
        for i in range(self.__table.rowCount()):
            name = self.__table.item(i, 0).text() if self.__table.item(i, 0) else ''
            value = self.__table.item(i, 1).text() if self.__table.item(i, 1) else ''
            if value.strip():
                prompt_text += f'{name}: {value}\n'
        return prompt_text

    def __generatePropPrompt(self, item: QTableWidgetItem):
        prompt_text = self.getPromptText()
        self.updated.emit(prompt_text)

    def __saveChangedPropPrompt(self, item: QTableWidgetItem):
        name = self.__table.item(item.row(), 0)
        id = name.data(Qt.UserRole)
        name = name.text()
        value = self.__table.item(item.row(), 1).text()
        self.__db.updatePropPromptAttribute(self.__id, id, name, value)

    def __add(self):
        dialog = InputDialog('Name', '', self)
        reply = dialog.exec()
        if reply == QDialog.Accepted:
            self.__table.itemChanged.disconnect(self.__saveChangedPropPrompt)

            name = dialog.getText()
            self.__table.setRowCount(self.__table.rowCount()+1)

            item1 = QTableWidgetItem(name)
            item1.setTextAlignment(Qt.AlignCenter)
            self.__table.setItem(self.__table.rowCount()-1, 0, item1)

            item2 = QTableWidgetItem('')
            item2.setTextAlignment(Qt.AlignCenter)
            self.__table.setItem(self.__table.rowCount()-1, 1, item2)

            id = self.__db.insertPropPromptAttribute(self.__id, name)
            item1.setData(Qt.UserRole, id)

            self.__table.itemChanged.connect(self.__saveChangedPropPrompt)

    def __delete(self):
        for i in sorted(set([i.row() for i in self.__table.selectedIndexes()]), reverse=True):
            id = self.__table.item(i, 0).data(Qt.UserRole)
            self.__table.removeRow(i)
            self.__db.deletePropPromptAttribute(self.__id, id)


class PropPage(QWidget):
    updated = Signal(str)

    def __init__(self, db: SqliteDatabase):
        super().__init__()
        self.__initVal(db)
        self.__initUi()

    def __initVal(self, db):
        self.__db = db
        self.__previousPropGroups = self.__db.selectPropPromptGroup()

    def __initUi(self):
        leftWidget = PropGroupList(self.__db)
        leftWidget.added.connect(self.__propGroupAdded)
        leftWidget.deleted.connect(self.__propGroupDeleted)
        leftWidget.currentRowChanged.connect(self.__showProp)

        self.__rightWidget = QStackedWidget()

        for group in self.__previousPropGroups:
            propTable = PropTable(self.__db, id=group[0])
            propTable.updated.connect(self.updated)
            self.__rightWidget.addWidget(propTable)

        mainWidget = QSplitter()
        mainWidget.addWidget(leftWidget)
        mainWidget.addWidget(self.__rightWidget)
        mainWidget.setChildrenCollapsible(False)
        mainWidget.setSizes([300, 700])

        lay = QVBoxLayout()
        lay.addWidget(mainWidget)

        self.setLayout(lay)

    def __propGroupAdded(self, id):
        propTable = PropTable(self.__db, id)
        propTable.updated.connect(self.updated)
        self.__rightWidget.addWidget(propTable)
        self.__rightWidget.setCurrentWidget(propTable)

    def __propGroupDeleted(self, n):
        w = self.__rightWidget.widget(n)
        self.__rightWidget.removeWidget(w)

    def __showProp(self, n):
        self.__rightWidget.setCurrentIndex(n)
        w = self.__rightWidget.currentWidget()
        if w and isinstance(w, PropTable):
            self.updated.emit(w.getPromptText())



from qtpy.QtWidgets import QWidget, QLineEdit, QGridLayout, QLabel, \
    QHBoxLayout, QApplication, QSizePolicy
from qtpy.QtCore import Signal
from pyqt_openai.svgLabel import SvgLabel


class SearchBar(QWidget):
    searched = Signal(str)

    def __init__(self, parent=None):
        super().__init__(parent)
        # search bar label
        self.__label = QLabel()

        self._initUi()

    def _initUi(self):
        self.__searchLineEdit = QLineEdit()
        self.__searchIconLbl = SvgLabel()
        ps = QApplication.font().pointSize()
        self.__searchIconLbl.setFixedSize(ps, ps)

        self.__searchBar = QWidget()
        self.__searchBar.setObjectName('searchBar')

        lay = QHBoxLayout()
        lay.addWidget(self.__searchIconLbl)
        lay.addWidget(self.__searchLineEdit)
        self.__searchBar.setLayout(lay)
        lay.setContentsMargins(ps // 2, 0, 0, 0)
        lay.setSpacing(0)

        self.__searchLineEdit.setFocus()
        self.__searchLineEdit.textChanged.connect(self.__searched)

        self.setAutoFillBackground(True)

        lay = QHBoxLayout()
        lay.addWidget(self.__searchBar)
        lay.setContentsMargins(0, 0, 0, 0)
        lay.setSpacing(2)

        self._topWidget = QWidget()
        self._topWidget.setLayout(lay)

        lay = QGridLayout()
        lay.addWidget(self._topWidget)

        searchWidget = QWidget()
        searchWidget.setLayout(lay)
        lay.setContentsMargins(0, 0, 0, 0)

        lay = QGridLayout()
        lay.addWidget(searchWidget)
        lay.setContentsMargins(0, 0, 0, 0)

        self.setSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.Preferred)

        self.__setStyle()

        self.setLayout(lay)

    # ex) searchBar.setLabel(True, 'Search Text')
    def setLabel(self, visibility: bool = True, text=None):
        if text:
            self.__label.setText(text)
        self.__label.setVisible(visibility)

    def __setStyle(self):
        self.__searchIconLbl.setSvgFile('ico/search.svg')
        self.setStyleSheet(
            '''
            QLineEdit
            {
                background: transparent;
                color:  #333333;
                border: none;
            }
            QWidget#searchBar
            {
                border: 1px solid gray;
            }
            QWidget { padding: 5px; }
            '''
        )

    def __searched(self, text):
        self.searched.emit(text)

    def setSearchIcon(self, icon_filename: str):
        self.__searchIconLbl.setIcon(icon_filename)

    def setPlaceHolder(self, text: str):
        self.__searchLineEdit.setPlaceholderText(text)

    def getSearchBar(self):
        return self.__searchLineEdit

    def getSearchLabel(self):
        return self.__searchIconLbl